from relationalai.early_access.metamodel.compiler import Pass
from relationalai.early_access.metamodel import ir, builtins as rel_builtins, factory as f, visitor
from relationalai.early_access.metamodel.typer import Checker, InferTypes, typer
from relationalai.early_access.metamodel import helpers, types
from relationalai.early_access.metamodel.rewrite import Splinter, ExtractNestedLogicals, ExtractKeys
from relationalai.early_access.metamodel.util import FrozenOrderedSet

from relationalai.early_access.metamodel.rewrite import Flatten
# TODO: Move this into metamodel.rewrite
from relationalai.early_access.rel.rewrite import QuantifyVars, CDC

from relationalai.early_access.lqp.utils import output_names

from typing import cast, List, Sequence, Tuple, Union, Optional, Iterable
from collections import defaultdict

def lqp_passes() -> list[Pass]:
    return [
        Checker(),
        ExtractNestedLogicals(), # before InferTypes to avoid extracting casts
        InferTypes(),
        CDC(),
        ExtractKeys(),
        # Broken
        # ExtractCommon(),
        Flatten(),
        Splinter(), # Splits multi-headed rules into multiple rules
        QuantifyVars(), # Adds missing existentials
        EliminateData(),  # Turns Data nodes into ordinary relations.
        DeduplicateVars(),  # Deduplicates vars in Updates and Outputs.
        PeriodMath(),  # Rewrite date period uses.
        ConstantsToVars(),  # Turns constants in Updates and Outputs into vars.
        UnifyDefinitions(),
    ]

# LQP does not support multiple definitions for the same relation. This pass unifies all
# definitions for each relation into a single definition using a union.
class UnifyDefinitions(Pass):
    def __init__(self):
        super().__init__()
        self.renamed_relation_args: dict[ir.Relation, list[ir.Var]] = {}

    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        root = cast(ir.Logical, model.root)
        new_tasks = self.get_combined_multidefs(root)
        return ir.Model(
            model.engines,
            model.relations,
            model.types,
            f.logical(
                tuple(new_tasks),
                root.hoisted,
                root.engine,
            ),
            model.annotations,
        )

    def _get_derives(self, logical: ir.Logical) -> list[ir.Update]:
        derives = []
        for task in logical.body:
            if isinstance(task, ir.Update) and task.effect == ir.Effect.derive:
                derives.append(task)
        return derives

    def _get_non_derives(self, logical: ir.Logical) -> list[ir.Update]:
        non_derives = []
        for task in logical.body:
            if not(isinstance(task, ir.Update) and task.effect == ir.Effect.derive):
                non_derives.append(task)
        return non_derives

    def get_combined_multidefs(self, root: ir.Logical) -> list[ir.Logical]:
        # Step 1: Group tasks by the relation they define.
        relation_to_tasks: dict[Optional[ir.Relation], list[ir.Logical]] = defaultdict(list)

        for task in root.body:
            task = cast(ir.Logical, task)
            task_derives = self._get_derives(task)

            # Some relations do not need to be grouped, e.g., if they don't contain a
            # derive. Use `None` as a placeholder key for these cases.
            if len(task_derives) != 1:
                relation_to_tasks[None].append(task)
                continue

            derive_name = task_derives[0].relation
            relation_to_tasks[derive_name].append(task)

        # Step 2: For each relation, combine all of the body definitions into a union.
        result_tasks = []
        for relation, tasks in relation_to_tasks.items():
            # If there's only one task for the relation, or if grouping is not needed, then
            # just keep the original tasks.
            if len(tasks) == 1 or relation is None:
                result_tasks.extend(tasks)
                continue

            result_tasks.append(self._combine_tasks_into_union(tasks))
        return result_tasks

    def _get_variable_mapping(self, logical: ir.Logical) -> dict[ir.Value, ir.Var]:
        derives = self._get_derives(logical)
        assert len(derives) == 1, "should only have one derive in a logical at this stage"
        derive = derives[0]

        var_mapping = {}

        if derive.relation not in self.renamed_relation_args:
            renamed_vars = []
            for (i, arg) in enumerate(derive.args):
                typ = typer.to_type(arg)
                assert arg not in var_mapping, "args of update should be unique"
                if isinstance(arg, ir.Var):
                    var_mapping[arg] = ir.Var(typ, arg.name)
                else:
                    var_mapping[arg] = ir.Var(typ, f"arg_{i}")

                renamed_vars.append(var_mapping[arg])

            self.renamed_relation_args[derive.relation] = renamed_vars
        else:
            for (arg, var) in zip(derive.args, self.renamed_relation_args[derive.relation]):
                var_mapping[arg] = var

        return var_mapping

    def _rename_variables(self, logical: ir.Logical) -> ir.Logical:
        class RenameVisitor(visitor.Rewriter):
            def __init__(self, var_mapping: dict[ir.Value, ir.Var]):
                super().__init__()
                self.var_mapping = var_mapping

            def _get_mapped_value(self, val: ir.Value) -> ir.Value:
                if isinstance(val, tuple):
                    return tuple(self._get_mapped_value(t) for t in val)
                return self.var_mapping.get(val, val)

            def _get_mapped_values(self, vals: Iterable[ir.Value]) -> list[ir.Value]:
                return [self._get_mapped_value(v) for v in vals]

            def handle_var(self, node: ir.Var, parent: ir.Node) -> ir.Var:
                return self.var_mapping.get(node, node)

            # TODO: ideally, extend the rewriter class to allow rewriting PyValue to Var so
            # we don't need to separately handle all cases containing them.
            def handle_update(self, node: ir.Update, parent: ir.Node) -> ir.Update:
                return ir.Update(
                    node.engine,
                    node.relation,
                    tuple(self._get_mapped_values(node.args)),
                    node.effect,
                    node.annotations,
                )

            def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
                return ir.Lookup(
                    node.engine,
                    node.relation,
                    tuple(self._get_mapped_values(node.args)),
                    node.annotations,
                )

            def handle_output(self, node: ir.Output, parent: ir.Node) -> ir.Output:
                new_aliases = FrozenOrderedSet(
                    [(name, self._get_mapped_value(value)) for name, value in node.aliases]
                )
                if node.keys:
                    new_keys = FrozenOrderedSet(
                        [self.var_mapping.get(key, key) for key in node.keys]
                    )
                else:
                    new_keys = node.keys

                return ir.Output(
                    node.engine,
                    new_aliases,
                    new_keys,
                    node.annotations,
                )

            def handle_construct(self, node: ir.Construct, parent: ir.Node) -> ir.Construct:
                new_values = tuple(self._get_mapped_values(node.values))
                new_id_var = self.var_mapping.get(node.id_var, node.id_var)
                return ir.Construct(
                    node.engine,
                    new_values,
                    new_id_var,
                    node.annotations,
                )

            def handle_aggregate(self, node: ir.Aggregate, parent: ir.Node) -> ir.Aggregate:
                new_projection = tuple(self.var_mapping.get(arg, arg) for arg in node.projection)
                new_group = tuple(self.var_mapping.get(arg, arg) for arg in node.group)
                new_args = tuple(self._get_mapped_values(node.args))
                return ir.Aggregate(
                    node.engine,
                    node.aggregation,
                    new_projection,
                    new_group,
                    new_args,
                    node.annotations,
                )

            def handle_rank(self, node: ir.Rank, parent: ir.Node) -> ir.Rank:
                new_projection = tuple(self.var_mapping.get(arg, arg) for arg in node.projection)
                new_group = tuple(self.var_mapping.get(arg, arg) for arg in node.group)
                new_args = tuple(self.var_mapping.get(arg, arg) for arg in node.args)
                new_result = self.var_mapping.get(node.result, node.result)

                return ir.Rank(
                    node.engine,
                    new_projection,
                    new_group,
                    new_args,
                    node.arg_is_ascending,
                    new_result,
                    node.limit,
                    node.annotations,
                )

        var_mapping = self._get_variable_mapping(logical)

        renamer = RenameVisitor(var_mapping)
        result = renamer.walk(logical)

        # Also need to append the equality for each renamed constant. E.g., if the mapping
        # contains (50.0::FLOAT -> arg_2::FLOAT), we need to add
        # `eq(arg_2::FLOAT, 50.0::FLOAT)` to the result.
        value_eqs = []
        for (old_var, new_var) in var_mapping.items():
            if not isinstance(old_var, ir.Var):
                value_eqs.append(f.lookup(rel_builtins.eq, [new_var, old_var]))

        return ir.Logical(
            result.engine,
            result.hoisted,
            tuple(value_eqs) + tuple(result.body),
            result.annotations,
        )

    # This function is the main workhorse for this rewrite pass. It takes a list of tasks
    # that define the same relation, and combines them into a single task that defines
    # the relation using a union of all of the bodies.
    def _combine_tasks_into_union(self, tasks: list[ir.Logical]) -> ir.Logical:
        # Step 1: Rename the variables in all tasks so that they will match the final derive
        # after reconstructing into a union
        renamed_tasks = [self._rename_variables(task) for task in tasks]

        # Step 2: Get the final derive
        derives = self._get_derives(renamed_tasks[0])
        assert len(derives) == 1, "should only have one derive in a logical at this stage"
        # Also make sure that all the derives are the same. This should be the case because
        # we renamed all the variables to be the same in step 1.
        for task in renamed_tasks[1:]:
            assert self._get_derives(task) == derives, "all derives should be the same"

        derive = derives[0]

        # Step 3: Remove the final `derive` from each task
        renamed_task_bodies = [
            f.logical(
                tuple(self._get_non_derives(t)),  # Only keep non-derive tasks
                t.hoisted,
                t.engine,
            )
            for t in renamed_tasks
        ]

        # Step 4: Construct a union of all the task bodies
        union = f.union(
            tuple(renamed_task_bodies),
            [],
            renamed_tasks[0].engine,
        )

        # Step 5: Add the final derive back
        return f.logical(
            (union, derive),
            [],
            renamed_tasks[0].engine,
        )

# Creates intermediary relations for all Data nodes and replaces said Data nodes
# with a Lookup into these created relations.
class EliminateData(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.DataRewriter()
        return r.walk(model)

    # Does the actual work.
    class DataRewriter(visitor.Rewriter):
        new_relations: list[ir.Relation]
        new_updates: list[ir.Logical]
        # Counter for naming new relations.
        # It must be that new_count == len new_updates == len new_relations.
        new_count: int

        def __init__(self):
            self.new_relations = []
            self.new_updates = []
            self.new_count = 0
            super().__init__()

        # Create a new intermediary relation representing the Data (and pop it in
        # new_updates/new_relations) and replace this Data with a Lookup of said
        # intermediary.
        def handle_data(self, node: ir.Data, parent: ir.Node) -> ir.Lookup:
            self.new_count += 1
            intermediary_name = f"formerly_Data_{self.new_count}"

            intermediary_relation = f.relation(
                intermediary_name,
                [f.field(v.name, v.type) for v in node.vars]
            )
            self.new_relations.append(intermediary_relation)

            intermediary_update = f.logical([
                # For each row (union), equate values and their variable (logical).
                f.union(
                    [
                        f.logical(
                            [
                                f.lookup(rel_builtins.eq, [f.literal(val), var])
                                for (val, var) in zip(row, node.vars)
                            ],
                            hoisted = node.vars,
                        )
                        for row in node
                    ],
                    hoisted = node.vars,
                ),
                # And pop it back into the relation.
                f.update(intermediary_relation, node.vars, ir.Effect.derive),
            ])
            self.new_updates.append(intermediary_update)

            replacement_lookup = f.lookup(intermediary_relation, node.vars)

            return replacement_lookup

        # Walks the model for the handle_data work then updates the model with
        # the new state.
        def handle_model(self, model: ir.Model, parent: None):
            walked_model = super().handle_model(model, parent)
            assert len(self.new_relations) == len(self.new_updates) and self.new_count == len(self.new_relations)

            # This is okay because its LQP.
            assert isinstance(walked_model.root, ir.Logical)
            root_logical = cast(ir.Logical, walked_model.root)

            # We may need to add the new intermediaries from handle_data to the model.
            if self.new_count  == 0:
                return model
            else:
                return ir.Model(
                    walked_model.engines,
                    walked_model.relations | self.new_relations,
                    walked_model.types,
                    ir.Logical(
                        root_logical.engine,
                        root_logical.hoisted,
                        root_logical.body + tuple(self.new_updates),
                        root_logical.annotations,
                    ),
                    walked_model.annotations,
                )

# Deduplicate Vars in Updates and Outputs.
class DeduplicateVars(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.VarDeduplicator()
        return r.walk(model)

    # Return 1) a new list of Values with no duplicates (at the object level) and
    # 2) equalities between any original Value and a deduplicated Value.
    @staticmethod
    def dedup_values(vals: Sequence[ir.Value]) -> Tuple[List[ir.Value], List[ir.Lookup]]:
        # If a var is seen more than once, it is a duplicate and we will create
        # a new Var and equate it with the seen one.
        seen_vars = set()

        new_vals = []
        eqs = []

        for i, val in enumerate(vals):
            # Duplicates can only occur within Vars.
            # TODO: we don't know for sure if these are the only relevant cases.
            if isinstance(val, ir.Default) or isinstance(val, ir.Var):
                var = val if isinstance(val, ir.Var) else val.var
                if var in seen_vars:
                    new_var = ir.Var(var.type, var.name + "_dup_" + str(i))
                    new_val = new_var if isinstance(val, ir.Var) else ir.Default(new_var, val.value)
                    new_vals.append(new_val)
                    eqs.append(f.lookup(rel_builtins.eq, [new_var, var]))
                else:
                    seen_vars.add(var)
                    new_vals.append(val)
            else:
                # No possibility of problematic duplication.
                new_vals.append(val)

        return new_vals, eqs

    # Returns a reconstructed output with no duplicate variable objects
    # (dedup_values) and now necessary equalities between any two previously
    # duplicate variables.
    @staticmethod
    def dedup_output(output: ir.Output) -> List[Union[ir.Output, ir.Lookup]]:
        vals = helpers.output_values(output.aliases)
        deduped_vals, req_lookups = DeduplicateVars.dedup_values(vals)
        # Need the names so we can recombine.
        alias_names = output_names(output.aliases)
        new_output = ir.Output(
            output.engine,
            FrozenOrderedSet(list(zip(alias_names, deduped_vals))),
            output.keys,
            output.annotations,
        )
        return req_lookups + [new_output]

    # Returns a replacement update with no duplicate variable objects
    # (dedup_values) and now necessary equalities between any two previously
    # duplicate variables.
    @staticmethod
    def dedup_update(update: ir.Update) -> List[Union[ir.Update, ir.Lookup]]:
        deduped_vals, req_lookups = DeduplicateVars.dedup_values(update.args)
        new_update = ir.Update(
            update.engine,
            update.relation,
            tuple(deduped_vals),
            update.effect,
            update.annotations,
        )
        return req_lookups + [new_update]

    # Does the actual work.
    class VarDeduplicator(visitor.Rewriter):
        def __init__(self):
            super().__init__()

        # We implement handle_logical instead of handle_update/handle_output
        # because in addition to modifying said update/output we require new
        # lookups (equality between original and deduplicated variables).
        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # In order to recurse over subtasks.
            node = super().handle_logical(node, parent)

            new_body = []
            for subtask in node.body:
                if isinstance(subtask, ir.Output):
                    new_body.extend(DeduplicateVars.dedup_output(subtask))
                elif isinstance(subtask, ir.Update):
                    new_body.extend(DeduplicateVars.dedup_update(subtask))
                else:
                    new_body.append(subtask)

            return ir.Logical(
                node.engine,
                node.hoisted,
                tuple(new_body),
                node.annotations
            )

# Generate date arithmetic expressions, such as
# `rel_primitive_date_add(:day, [date] delta, res_2)` by finding the period
# expression for the delta and adding the period type to the date arithmetic expression.
#
# date_add and it's kin are generated by a period expression, e.g.,
# `day(delta, res_1)`
# followed by the date arithmetic expression using the period
# `date_add([date] res_1 res_2)`
class PeriodMath(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        period_rewriter = self.PeriodRewriter()
        model = period_rewriter.walk(model)
        period_math_rewriter = self.PeriodMathRewriter(period_rewriter.period_vars)
        model = period_math_rewriter.walk(model)
        return model

    # Find all period builtins. We need to make them safe for the emitter (either by
    # translating to a cast, or removing) and store the variable and period type for use
    # in the date/datetime add/subtract expressions.
    class PeriodRewriter(visitor.Rewriter):
        def __init__(self):
            super().__init__()
            self.period_vars: dict[ir.Var, str] = {}

        def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
            if not rel_builtins.is_builtin(node.relation):
                return node

            if node.relation.name not in {
                "year", "month", "week", "day", "hour", "minute", "second", "millisecond"
            }:
                return node

            assert len(node.args) == 2, "Expect 2 arguments for period builtins"
            assert isinstance(node.args[1], ir.Var), "Expect result to be a variable"
            period = node.relation.name
            result_var = node.args[1]
            self.period_vars[result_var] = period

            # Ideally we could now remove the unused and unhandled period type construction
            # but we may also need to cast the original variable to an Int64 for use by the
            # date/datetime add/subtract expressions.
            # TODO: Remove the node entirely where possible and update uses of the result
            return f.lookup(rel_builtins.cast, [types.Int64, node.args[0], result_var])

    # Update date/datetime add/subtract expressions with period information.
    class PeriodMathRewriter(visitor.Rewriter):
        def __init__(self, period_vars: dict[ir.Var, str]):
            super().__init__()
            self.period_vars: dict[ir.Var, str] = period_vars

        def handle_lookup(self, node: ir.Lookup, parent: ir.Node) -> ir.Lookup:
            if not rel_builtins.is_builtin(node.relation):
                return node

            if node.relation.name not in {
                "date_add", "date_subtract", "datetime_add", "datetime_subtract"
            }:
                return node

            if len(node.args) == 4:
                # We've already visited this lookup
                return node

            assert isinstance(node.args[1], ir.Var), "Expect period to be a variable"
            period_var = node.args[1]
            assert period_var in self.period_vars, "datemath found, but no vars to insert"

            period = self.period_vars[period_var]

            new_args = [f.literal(period, types.Symbol)] + [arg for arg in node.args]

            return f.lookup(node.relation, new_args)

# Rewrite constants to vars in Updates. This results in a more normalized format where
# updates contain only variables. This allows for easier rewrites in later passes.
class ConstantsToVars(Pass):
    def rewrite(self, model: ir.Model, options:dict={}) -> ir.Model:
        r = self.ConstantToVarRewriter()
        return r.walk(model)

    # Return 1) a new list of Values with no duplicates (at the object level) and
    # 2) equalities between any original Value and a deduplicated Value.
    @staticmethod
    def replace_constants_with_vars(vals: Sequence[ir.Value]) -> Tuple[List[ir.Value], List[ir.Lookup]]:
        new_vals = []
        eqs = []

        for i, val in enumerate(vals):
            if isinstance(val, ir.PyValue) or isinstance(val, ir.Literal):
                # Replace constant with a new Var.
                typ = typer.to_type(val)
                assert isinstance(typ, ir.ScalarType), "can only replace scalar constants with vars"
                new_var = ir.Var(typ, f"{typ.name.lower()}_{i}")
                new_vals.append(new_var)
                eqs.append(f.lookup(rel_builtins.eq, [new_var, val]))
            else:
                new_vals.append(val)

        return new_vals, eqs

    @staticmethod
    def dedup_update(update: ir.Update) -> List[Union[ir.Update, ir.Lookup]]:
        deduped_vals, req_lookups = ConstantsToVars.replace_constants_with_vars(update.args)
        new_update = ir.Update(
            update.engine,
            update.relation,
            tuple(deduped_vals),
            update.effect,
            update.annotations,
        )
        return req_lookups + [new_update]

    # Does the actual work.
    class ConstantToVarRewriter(visitor.Rewriter):
        def __init__(self):
            super().__init__()

        # We implement handle_logical instead of handle_update because in
        # addition to modifying said update we require new lookups (equality
        # between original and deduplicated variables).
        def handle_logical(self, node: ir.Logical, parent: ir.Node):
            # In order to recurse over subtasks.
            node = super().handle_logical(node, parent)

            new_body = []
            for subtask in node.body:
                if isinstance(subtask, ir.Update):
                    new_body.extend(ConstantsToVars.dedup_update(subtask))
                else:
                    new_body.append(subtask)

            return ir.Logical(
                node.engine,
                node.hoisted,
                tuple(new_body),
                node.annotations
            )



from __future__ import annotations
import atexit
from collections import defaultdict
import json
import re
import textwrap
import uuid

from pandas import DataFrame
from typing import Any, Optional
import relationalai as rai

from relationalai import debugging
from relationalai.clients import result_helpers
from relationalai.clients.util import IdentityParser
from relationalai.clients.snowflake import APP_NAME
from relationalai.early_access.metamodel import ir, executor as e, factory as f
from relationalai.early_access.rel import Compiler
from relationalai.clients.config import Config
from relationalai.tools.constants import USE_DIRECT_ACCESS, Generation


class RelExecutor(e.Executor):
    """Executes Rel code using the RAI client."""

    def __init__(
        self,
        database: str,
        dry_run: bool = False,
        keep_model: bool = True,
        wide_outputs: bool = False,
        config: Config | None = None,
    ) -> None:
        super().__init__()
        self.database = database
        self.dry_run = dry_run
        self.keep_model = keep_model
        self.wide_outputs = wide_outputs
        self.compiler = Compiler()
        self.config = config or Config()
        self._resources = None
        self._last_model = None
        self._last_sources_version = (-1, None)

    @property
    def resources(self):
        if not self._resources:
            with debugging.span("create_session"):
                self.dry_run |= bool(self.config.get("compiler.dry_run", False))
                resource_class = rai.clients.snowflake.Resources
                if self.config.get("use_direct_access", USE_DIRECT_ACCESS):
                    resource_class = rai.clients.snowflake.DirectAccessResources
                self._resources = resource_class(dry_run=self.dry_run, config=self.config, generation=rai.Generation.QB)
                if not self.dry_run:
                    self.engine = self._resources.get_default_engine_name()
                    if not self.keep_model:
                        atexit.register(self._resources.delete_graph, self.database, True)
        return self._resources

    def check_graph_index(self):
        # Has to happen first, so self.dry_run is populated.
        resources = self.resources

        if self.dry_run:
            return

        from relationalai.early_access.builder.snowflake import Table
        table_sources = Table._used_sources
        if not table_sources.has_changed(self._last_sources_version):
            return

        model = self.database
        app_name = resources.get_app_name()
        engine_name = self.engine
        engine_size = self.resources.config.get_default_engine_size()

        program_span_id = debugging.get_program_span_id()
        sources = [t._fqn for t in Table._used_sources]
        self._last_sources_version = Table._used_sources.version()

        assert self.engine is not None

        with debugging.span("poll_use_index", sources=sources, model=model, engine=engine_name):
            resources.poll_use_index(app_name, sources, model, self.engine, engine_size, program_span_id)

    def report_errors(self, problems: list[dict[str, Any]], abort_on_error=True):
        from relationalai import errors
        all_errors = []
        undefineds = []
        pyrel_errors = defaultdict(list)
        pyrel_warnings = defaultdict(list)

        for problem in problems:
            message = problem.get("message", "")
            report = problem.get("report", "")
            # TODO: we need to build source maps
            # path = problem.get("path", "")
            # source_task = self._install_batch.line_to_task(path, problem["start_line"]) or task
            # source = debugging.get_source(source_task) or debugging.SourceInfo()
            source = debugging.SourceInfo()
            severity = problem.get("severity", "warning")
            code = problem.get("code")

            if severity in ["error", "exception"]:
                if code == "UNDEFINED_IDENTIFIER":
                    match = re.search(r'`(.+?)` is undefined', message)
                    if match:
                        undefineds.append((match.group(1), source))
                    else:
                        all_errors.append(errors.RelQueryError(problem, source))
                elif "overflowed" in report:
                    all_errors.append(errors.NumericOverflow(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_errors[problem["props"]["pyrel_id"]].append(problem)
                elif abort_on_error:
                    all_errors.append(errors.RelQueryError(problem, source))
            else:
                if code == "ARITY_MISMATCH":
                    errors.ArityMismatch(problem, source)
                elif code == "IC_VIOLATION":
                    all_errors.append(errors.IntegrityConstraintViolation(problem, source))
                elif code == "PYREL_ERROR":
                    pyrel_warnings[problem["props"]["pyrel_id"]].append(problem)
                else:
                    errors.RelQueryWarning(problem, source)

        if abort_on_error and len(undefineds):
            all_errors.append(errors.UninitializedPropertyException(undefineds))

        if abort_on_error:
            for pyrel_id, pyrel_problems in pyrel_errors.items():
                all_errors.append(errors.ModelError(pyrel_problems))

        for pyrel_id, pyrel_problems in pyrel_warnings.items():
            errors.ModelWarning(pyrel_problems)


        if len(all_errors) == 1:
            raise all_errors[0]
        elif len(all_errors) > 1:
            raise errors.RAIExceptionSet(all_errors)

    def _export(self, raw_code: str, dest_fqn: str, actual_cols: list[str], declared_cols: list[str], update:bool):
        _exec = self.resources._exec
        output_table = "out" + str(uuid.uuid4()).replace("-", "_")
        txn_id = None
        artifacts = None
        dest_database, dest_schema, dest_table, _ = IdentityParser(dest_fqn, require_all_parts=True).to_list()
        assert self.resources._session
        with debugging.span("transaction"):
            try:
                with debugging.span("exec_format") as span:
                    res = _exec(f"call {APP_NAME}.api.exec_into_table(?, ?, ?, ?, ?, ?);", [self.database, self.engine, raw_code, output_table, False, True])
                    txn_id = json.loads(res[0]["EXEC_INTO_TABLE"])["rai_transaction_id"]
                    span["txn_id"] = txn_id

                with debugging.span("write_table", txn_id=txn_id):
                    out_sample = _exec(f"select * from {APP_NAME}.results.{output_table} limit 1;")
                    if out_sample:
                        keys = set([k.lower() for k in out_sample[0].as_dict().keys()])
                        fields = []
                        ix = 0
                        for name in declared_cols:
                            if name in actual_cols:
                                field = f"col{ix:03} as \"{name}\"" if f"col{ix:03}" in keys else f"NULL as {name}"
                                ix += 1
                            else:
                                field = f"NULL as \"{name}\""
                            fields.append(field)
                        names = ", ".join(fields)
                        if not update:
                            _exec(f"""
                                BEGIN
                                    -- Check if table exists
                                    IF (EXISTS (
                                        SELECT 1
                                        FROM {dest_database}.INFORMATION_SCHEMA.TABLES
                                        WHERE table_schema = '{dest_schema}'
                                        AND table_name = '{dest_table}'
                                    )) THEN
                                        -- Insert into existing table
                                        EXECUTE IMMEDIATE '
                                            BEGIN
                                                TRUNCATE TABLE {dest_fqn};
                                                INSERT INTO {dest_fqn}
                                                SELECT {names}
                                                FROM {APP_NAME}.results.{output_table};
                                            END;
                                        ';
                                    ELSE
                                        -- Create table based on the SELECT
                                        EXECUTE IMMEDIATE '
                                            CREATE TABLE {dest_fqn} AS
                                            SELECT {names}
                                            FROM {APP_NAME}.results.{output_table};
                                        ';
                                    END IF;
                                    CALL {APP_NAME}.api.drop_result_table('{output_table}');
                                END;
                            """)
                        else:
                            _exec(f"""
                                begin
                                    INSERT INTO {dest_fqn} SELECT {names} FROM {APP_NAME}.results.{output_table};
                                    call {APP_NAME}.api.drop_result_table('{output_table}');
                                end;
                            """)
            except Exception as e:
                msg = str(e).lower()
                if "no columns returned" in msg or "columns of results could not be determined" in msg:
                    pass
                else:
                    raise e
            if txn_id:
                artifact_info = self.resources._list_exec_async_artifacts(txn_id)
                with debugging.span("fetch"):
                    artifacts = self.resources._download_results(artifact_info, txn_id, "ABORTED")
            return artifacts

    def execute(self, model: ir.Model, task:ir.Task, result_cols:list[str]|None=None, export_to:Optional[str]=None, update:bool=False) -> DataFrame:
        self.check_graph_index()
        resources= self.resources

        rules_code = ""
        if self._last_model != model:
            with debugging.span("compile", metamodel=model) as install_span:
                base = textwrap.dedent("""
                    declare pyrel_error_attrs(err in ::std::common::UInt128, attr in ::std::common::String, v) requires true

                """)
                rules_code = base + self.compiler.compile(model, {"wide_outputs": self.wide_outputs})
                install_span["compile_type"] = "model"
                install_span["rel"] = rules_code
                rules_code = resources.create_models_code([("pyrel_qb_0", rules_code)])
                self._last_model = model


        with debugging.span("compile", metamodel=task) as compile_span:
            base = textwrap.dedent("""
                def output(:pyrel_error, err, attr, val):
                    pyrel_error_attrs(err, attr, val)

            """)
            task_model = f.compute_model(f.logical([task]))
            task_code, task_model = self.compiler.compile_inner(task_model, {"no_declares": True, "wide_outputs": self.wide_outputs})
            task_code = base + task_code
            compile_span["compile_type"] = "query"
            compile_span["rel"] = task_code

        full_code = textwrap.dedent(f"""
            {rules_code}
            {task_code}
        """)

        if self.dry_run:
            return DataFrame()

        cols, extra_cols = self._compute_cols(task, task_model)

        if not export_to:
            raw_results = resources.exec_raw(self.database, self.engine, full_code, False, nowait_durable=True)
            df, errs = result_helpers.format_results(raw_results, None, cols, generation=Generation.QB)  # Pass None for task parameter
            self.report_errors(errs)
            return self._postprocess_df(self.config, df, extra_cols)
        else:
            assert cols
            # The result cols should be a superset of the actual cols.
            if result_cols is not None:
                assert all(col in result_cols or col in extra_cols for col in cols)
            else:
                result_cols = [col for col in cols if col not in extra_cols]
            assert result_cols
            raw = self._export(full_code, export_to, cols, result_cols, update)
            errors = []
            if raw:
                dataframe, errors = result_helpers.format_results(raw, None, result_cols, generation=Generation.QB)
                self.report_errors(errors)
            return DataFrame()

    # NOTE(coey): this is added temporarily to support executing Rel for the solvers library in EA.
    # It can be removed once this is no longer needed by the solvers library.
    def execute_raw(self, raw_rel:str, readonly:bool=True) -> DataFrame:
        raw_results = self.resources.exec_raw(self.database, self.engine, raw_rel, readonly, nowait_durable=True)
        df, errs = result_helpers.format_results(raw_results, None, generation=Generation.QB)  # Pass None for task parameter
        self.report_errors(errs)
        return df



from __future__ import annotations

import re
import uuid

import relationalai as rai
import pandas as pd

from typing import Any, Union, cast, Optional

from .. import Compiler
from relationalai import debugging
from relationalai.early_access.metamodel import ir, executor as e, factory as f
from relationalai.clients.result_helpers import sort_data_frame_result
from relationalai.semantics.sql.executor.result_helpers import format_columns
from relationalai.early_access.metamodel.visitor import collect_by_type
from relationalai.early_access.metamodel.typer import typer

class SnowflakeExecutor(e.Executor):
    """Executes SQL using the RAI client."""

    def __init__(
            self,
            database: str,
            schema: str,
            dry_run: bool = False,
            skip_denormalization: bool = False,
            config: rai.Config | None = None,
    ) -> None:
        super().__init__()
        self.database = database
        self.schema = schema
        self.dry_run = dry_run
        self._last_model = None
        self._last_model_sql = None
        self.config = config or rai.Config()
        self.compiler = Compiler(skip_denormalization)
        self.provider = cast(rai.clients.snowflake.Provider, rai.Provider(config=config))

    def execute(self, model: ir.Model, task:ir.Task, result_cols:Optional[list[str]]=None,
                export_to:Optional[str]=None, update:bool=False) -> Union[pd.DataFrame, Any]:
        """ Execute the SQL query directly. """

        if self._last_model != model:
            with debugging.span("compile", metamodel=model) as model_span:
                model_sql, _ = self.compiler.compile(model)
                model_span["compile_type"] = "model"
                model_span["sql"] = model_sql
                self._last_model = model
                self._last_model_sql = model_sql

        with debugging.span("compile", metamodel=task) as compile_span:
            # compile into sql and keep the new_task, which is the task model after rewrites,
            # as it may contain results of type inference, which is useful for determining
            # how to format the outputs
            query_sql, new_task = self.compiler.compile(f.compute_model(task), {"query_compilation": True})
            compile_span["compile_type"] = "query"
            compile_span["sql"] = query_sql

        if self.dry_run:
            return pd.DataFrame()

        _replace_pattern = re.compile(r"[ /:-]")

        def sanitize_name(value: str) -> str:
            return _replace_pattern.sub("_", value)

        database = sanitize_name(self.database)
        schema = sanitize_name(self.schema)
        unique_id = sanitize_name(str(uuid.uuid4()).lower())
        db_name = f"{database}_{unique_id}"
        db_query = f"CREATE OR REPLACE DATABASE {db_name};"
        schema_query = f"CREATE OR REPLACE SCHEMA {db_name}.{schema};"
        use_schema_query = f"USE SCHEMA {db_name}.{schema};"

        full_model_sql = f"{db_query}\n{schema_query}\n{use_schema_query}\n{self._last_model_sql}\n{query_sql}"

        try:
            result = self.provider.resources._session.connection.execute_string(full_model_sql) # type: ignore

            # Assuming that `task` is a single SQL query per model, and we have it always at the end of the generated SQL.
            rows = result[-1].fetchall()
            result_metadata = result[-1].description

            df = pd.DataFrame(rows, columns=[col.name for col in result_metadata])
            if df.empty:
                # return empty df without column names if it's empty
                return df.iloc[:, 0:0]

            df = format_columns(df, result_metadata, self._collect_columns_metadata(new_task))
            return sort_data_frame_result(df)

        finally:
            self.provider.sql(f"DROP DATABASE IF EXISTS {db_name};")

    def _collect_columns_metadata(self, task: ir.Task) -> dict[str, Optional[ir.Type]]:
        if not task:
            return {}

        outputs = collect_by_type(ir.Output, task)

        assert len(outputs) == 1
        return {
            alias.lower(): typer.to_base_primitive(var.type)
            for alias, var in outputs[0].aliases
            if alias and isinstance(var, ir.Var)
        }



import warnings
from collections import defaultdict
from typing import Optional
from relationalai import Config
from relationalai.early_access.dsl.adapters.orm.model import ExclusiveInclusiveSubtypeFact, ExclusiveSubtypeFact, \
    InclusiveSubtypeFact, ORMRingType, ORMValueComparisonOperator, ORMRange
from relationalai.early_access.dsl.adapters.orm.parser import ORMParser
from relationalai.early_access.dsl.core.utils import to_pascal_case
from relationalai.early_access.dsl.orm.constraints import RingType, ValueComparisonType, Range
from relationalai.early_access.dsl.orm.models import Model
from relationalai.early_access.builder import Concept, Integer, String, DateTime, Decimal, Date, Float
from relationalai.early_access.metamodel.util import NameCache
from datetime import date, datetime
from decimal import Decimal as PyDecimal

class ORMAdapterQB:
    _datatype_mapping = {
        "SignedIntegerNumericDataType": Integer,
        "SignedLargeIntegerNumericDataType": Integer,
        "UnsignedIntegerNumericDataType": Integer,
        "UnsignedTinyIntegerNumericDataType": Integer,
        "UnsignedSmallIntegerNumericDataType": Integer,
        "UnsignedLargeIntegerNumericDataType": Integer,
        "AutoCounterNumericDataType": Integer,
        "FloatingPointNumericDataType": Float,
        "SinglePrecisionFloatingPointNumericDataType": Float,
        "DoublePrecisionFloatingPointNumericDataType": Float,
        "DecimalNumericDataType": Decimal,
        "MoneyNumericDataType": Decimal,
        "FixedLengthTextDataType": String,
        "VariableLengthTextDataType": String,
        "LargeLengthTextDataType": String,
        "DateTemporalDataType": Date,
        "DateAndTimeTemporalDataType": DateTime,
        "AutoTimestampTemporalDataType": DateTime,
        "TimeTemporalDataType": DateTime,
    }

    _builtin_concepts_to_python_types = {
        Concept.builtins["Int"]: int,
        Concept.builtins["Float"]: float,
        Concept.builtins["String"]: str,
        Concept.builtins["Bool"]: bool,
        Concept.builtins["Date"]: date,
        Concept.builtins["DateTime"]: datetime,
        Concept.builtins["Decimal"]: PyDecimal
    }

    _ring_type_mapping = {
        ORMRingType.IRREFLEXIVE: RingType.IRREFLEXIVE,
        ORMRingType.ANTISYMMETRIC: RingType.ANTISYMMETRIC,
        ORMRingType.ASYMMETRIC: RingType.ASYMMETRIC,
        ORMRingType.INTRANSITIVE: RingType.INTRANSITIVE,
        ORMRingType.STRONGLY_INTRANSITIVE: RingType.STRONGLY_INTRANSITIVE,
        ORMRingType.ACYCLIC: RingType.ACYCLIC,
        ORMRingType.PURELY_REFLEXIVE: RingType.PURELY_REFLEXIVE,
        ORMRingType.REFLEXIVE: RingType.REFLEXIVE,
        ORMRingType.SYMMETRIC: RingType.SYMMETRIC,
        ORMRingType.TRANSITIVE: RingType.TRANSITIVE
    }

    _value_comparison_type_mapping = {
        ORMValueComparisonOperator.GREATER_THAN_OR_EQUAL: ValueComparisonType.GREATER_THAN_OR_EQUAL,
        ORMValueComparisonOperator.LESS_THAN_OR_EQUAL: ValueComparisonType.LESS_THAN_OR_EQUAL,
        ORMValueComparisonOperator.GREATER_THAN: ValueComparisonType.GREATER_THAN,
        ORMValueComparisonOperator.LESS_THAN: ValueComparisonType.LESS_THAN,
        ORMValueComparisonOperator.NOT_EQUAL: ValueComparisonType.NOT_EQUAL,
        ORMValueComparisonOperator.EQUAL: ValueComparisonType.EQUAL
    }

    def __init__(self, orm_file_path: str, model_name: Optional[str] = None, config: Optional[Config] = None):
        self._parser = ORMParser(orm_file_path)
        self._relationship_role_value_constraints = defaultdict()
        self._relationships = {}
        self.name_cache = NameCache()
        self.model = self._orm_to_model(model_name, config)

    def _orm_to_model(self, model_name, config):
        model = Model(model_name, config=config) if model_name else Model(self._parser.model_name(), config=config)

        self._add_value_types(model)
        self._add_entity_types(model)
        self._add_subtype_relationships(model)
        self._add_relationships(model)
        self._add_external_identifying_relationships(model)
        self._add_role_value_constraints(model)
        self._add_inclusive_role_constraints(model)
        self._add_exclusive_role_constraints(model)
        self._add_ring_constraints(model)
        self._add_value_comparison_constraints(model)
        self._add_role_subset_constraints(model)
        self._add_equality_constraints(model)
        self._add_frequency_constraints(model)
        return model

    def _add_value_types(self, model):
        enum_to_values = defaultdict(list)
        # Value types having a value type constraint with only string values are considered Enums
        for v in self._parser.value_type_value_constraints().values():
            if self._is_value_type_an_enum(v.id):
                enum_to_values[v.id].extend(v.values)
        # Value types playing a role in only one relationship that has a role value constraint with only
        # string values are also considered Enums
        for rvc in self._parser.role_value_constraints().values():
            if self._is_value_type_playing_role_an_enum(rvc.role.id):
                if rvc.role.player not in enum_to_values.keys():
                    enum_to_values[rvc.role.player].extend(rvc.values)

        cardinality_constraints = self._parser.objects_type_to_cardinality_constraints()
        value_constraints = self._parser.value_type_value_constraints()
        for vt in self._parser.value_types().values():
            if vt.id in enum_to_values:
                concept = model.Enum(to_pascal_case(vt.name), enum_to_values[vt.id])
            else:
                concept = model.Concept(vt.name, extends=[self._datatype_mapping.get(vt.data_type, String)])
            if vt.name in cardinality_constraints:
                model.cardinality(concept, self._build_constraint_ranges(cardinality_constraints[vt.name], int))
            if vt.id in value_constraints:
                tp = self._builtin_concepts_to_python_types[concept._extends[0]]
                model.value_constraint(concept, self._build_constraint_ranges(value_constraints[vt.id], tp))

    def _add_entity_types(self, model):
        extended_concepts = [y.subtype_name for x in self._parser.subtype_facts().values() for y in x]
        cardinality_constraints = self._parser.objects_type_to_cardinality_constraints()
        for et in self._parser.entity_types().values():
            if et.name not in extended_concepts:
                concept = model.Concept(et.name)
                if et.name in cardinality_constraints:
                    model.cardinality(concept, self._build_constraint_ranges(cardinality_constraints[et.name], int))

        subtype_facts = self._parser.sorted_subtype_facts()
        for st_fact in subtype_facts:
            parent = st_fact.supertype_name
            parent_entity = model.lookup_concept(parent)
            if parent_entity is None:
                parent_entity = model.Concept(parent)
                if parent in cardinality_constraints:
                    model.cardinality(parent_entity, self._build_constraint_ranges(cardinality_constraints[parent], int))
            child = st_fact.subtype_name
            child_entity = model.Concept(child, extends=[parent_entity])
            if child in cardinality_constraints:
                model.cardinality(child_entity, self._build_constraint_ranges(cardinality_constraints[child], int))

    def _add_subtype_relationships(self, model):
        for parent, children in self._parser.subtype_facts().items():
            exclusive_subtypes = []
            inclusive_subtypes = []
            exclusive_inclusive_subtypes = []
            for child in children:
                sub = model.lookup_concept(child.subtype_name)
                if isinstance(child, ExclusiveInclusiveSubtypeFact):
                    exclusive_inclusive_subtypes.append(sub)
                elif isinstance(child, ExclusiveSubtypeFact):
                    exclusive_subtypes.append(sub)
                elif isinstance(child, InclusiveSubtypeFact):
                    inclusive_subtypes.append(sub)

            if len(exclusive_inclusive_subtypes) > 0:
                model.exclusive_subtype_constraint(*exclusive_inclusive_subtypes)
                model.inclusive_subtype_constraint(*exclusive_inclusive_subtypes)
            if len(exclusive_subtypes) > 0:
                model.exclusive_subtype_constraint(*exclusive_subtypes)
            if len(inclusive_subtypes) > 0:
                model.inclusive_subtype_constraint(*inclusive_subtypes)

    def _add_relationships(self, model):
        object_types = self._parser.object_types()
        unique_roles = self._parser.unique_roles()
        mandatory_roles = self._parser.mandatory_roles()
        role_value_constraints = self._parser.role_value_constraints()
        role_cardinality_constraints = self._parser.role_cardinality_constraints()
        fact_type_to_internal_ucs = self._parser.fact_type_to_internal_ucs()
        fact_type_to_complex_ucs = self._parser.fact_type_to_complex_ucs()
        fact_type_to_roles = self._parser.fact_type_to_roles()
        for fact_type, reading_orders in self._parser.fact_type_readings().items():

            # Adding the main reading
            rdo = reading_orders[0]
            player = object_types[rdo.roles[0].role.player].name
            player_entity = model.lookup_concept(player)
            reading = self._build_reading(model, rdo)
            relationship = model.Relationship(reading)
            self._set_role_verbalization(relationship, rdo)
            short_name = self._pick_short_name(model, player_entity, fact_type, relationship._readings[0])
            cached_name = self.name_cache.get_name(relationship._readings[0], f"{player_entity._name}.{short_name}")
            short_name = cached_name.split(".")[1]
            setattr(player_entity, short_name, relationship)
            self._relationships[fact_type] = relationship


            # Marking unique and mandatory roles
            role_idx_to_player = []
            for ro in rdo.roles:
                role_id = ro.role.id
                role_idx_to_player.append(role_id)
                role_index = role_idx_to_player.index(role_id)
                role = relationship[role_index]
                if role_id in unique_roles:
                    model.unique(role)
                if role_id in mandatory_roles:
                    model.mandatory(role)
                if role_id in role_value_constraints.keys():
                    self._relationship_role_value_constraints[role] = role_value_constraints[role_id].values
                if role_id in role_cardinality_constraints.keys():
                    model.role_cardinality(role, self._build_constraint_ranges(role_cardinality_constraints[role_id], int))


            # Adding alternative readings
            self._validate_role_verbalization(reading_orders)

            if len(reading_orders) > 1:
                for rdo in reading_orders[1:]:
                    other_player = object_types[rdo.roles[0].role.player].name
                    other_player_entity = model.lookup_concept(other_player)
                    alt_reading = self._build_reading(model, rdo)
                    alt_reading_obj = relationship.alt(alt_reading)
                    short_name = self._pick_short_name(model, other_player_entity, fact_type, alt_reading_obj)
                    cached_name = self.name_cache.get_name(alt_reading_obj, f"{other_player_entity._name}.{short_name}")
                    short_name = cached_name.split(".")[1]
                    setattr(other_player_entity, short_name, alt_reading_obj)

            # Marking identifying relationships
            if fact_type in fact_type_to_internal_ucs:
                for uc in fact_type_to_internal_ucs[fact_type]:
                    if uc.identifies:
                        player_entity.identify_by(relationship)

            # Adding constraint spanning over multiple roles
            if fact_type in fact_type_to_complex_ucs:
                for uc in fact_type_to_complex_ucs[fact_type]:
                    uc_roles = []
                    for role in fact_type_to_roles[fact_type]:

                        if role.id in uc.roles:
                            p = role.id
                            rl = role_idx_to_player.index(p)
                            uc_roles.append(self._relationships[fact_type][rl])
                    model.unique(*uc_roles)

    def _add_external_identifying_relationships(self, model):
        roles = self._parser.roles()
        object_types = self._parser.object_types()
        for uc in self._parser.external_uniqueness_constraints().values():
            # Identifying external UCs
            if uc.identifies:
                entity = model.lookup_concept(object_types[uc.identifies].name)
                identifying_relationships = []
                for ro in uc.roles:
                    relationship = self._relationships[roles[ro].relationship_name]
                    first_player = relationship._roles()[0]._concept
                    if first_player._name == entity._name:
                        identifying_relationships.append(relationship)
                    else:
                        alt_id_reading = self._get_alternative_identifying_reading(entity, relationship)
                        if alt_id_reading:
                            identifying_relationships.append(alt_id_reading)
                        else:
                            # Unless there exists an alternative reading where the identified entity is the first player,
                            # we cannot use this relationship as identifying. The issue is that without an explicit
                            # reading, we have no way to name and index the identifying relationship which in turn
                            # needs to be passed to qb identify_by.
                            # See also: https://relationalai.atlassian.net/browse/RAI-39727.
                            raise ValueError(f"The identifying relationship {first_player._name}.{relationship._name} must have a reading where the identified entity is the first player.")
                entity.identify_by(*identifying_relationships)
            # Non identifying external UCs
            else:
                role_list = []
                for ro in uc.roles:
                    relationship = self._relationships[roles[ro].relationship_name]
                    role = relationship._rel_roles[object_types[roles[ro].player].name.lower()]
                    role_list.append(role)
                model.unique(*role_list)

    def _add_inclusive_role_constraints(self, model):
        for rc in self._parser.inclusive_role_constraints():
            constraint_roles = self._get_roles_from_orm_constraint(model, rc)
            model.inclusive_roles(*constraint_roles)

    def _add_exclusive_role_constraints(self, model):
        for rc in self._parser.exclusive_role_constraints():
            constraint_roles = self._get_roles_from_orm_constraint(model, rc)
            model.exclusive_roles(*constraint_roles)

    def _add_ring_constraints(self, model):
        for rc in self._parser.ring_constraints().values():
            constraint_roles = self._get_roles_from_orm_constraint(model, rc)
            constraint_types = [self._ring_type_mapping.get(rt) for rt in rc.ring_types]
            model.ring(constraint_types, *constraint_roles)

    def _add_value_comparison_constraints(self, model):
        for vcc in self._parser.value_comparison_constraints().values():
            constraint_roles = self._get_roles_from_orm_constraint(model, vcc)
            constraint_type = self._value_comparison_type_mapping.get(vcc.operator)
            model.value_comparison(constraint_type, *constraint_roles)

    def _add_role_subset_constraints(self, model):
        for rsc in self._parser.role_subset_constraints().values():
            constraint_roles = self._get_roles_from_orm_constraint(model, rsc)
            model.role_subset(*constraint_roles)

    def _add_equality_constraints(self, model):
        for ec in self._parser.equality_constraints().values():
            constraint_roles = self._get_roles_from_orm_constraint(model, ec)
            model.equality(*constraint_roles)

    def _add_frequency_constraints(self, model):
        for fc in self._parser.frequency_constraints().values():
            constraint_roles = self._get_roles_from_orm_constraint(model, fc)
            constraint_frequency = (fc.min_frequency, fc.max_frequency)
            model.frequency(constraint_frequency, *constraint_roles)

    def _add_role_value_constraints(self, model):
        for role, values in self._relationship_role_value_constraints.items():
            extends = role._concept._extends[0]
            # If the value type was created as an Enum, we can use the values directly
            if extends._name == "Enum":
                model.role_value_constraint(role, values)
            # Otherwise we will process a mix of single values and ORMRange objects
            else:
                tp = self._builtin_concepts_to_python_types.get(extends)
                if tp:
                    constraint_values = []
                    for value in values:
                        if isinstance(value, ORMRange):
                            constraint_values.append(self._build_constraint_range(value, tp))
                        else:
                            constraint_values.append(tp(value))
                    model.role_value_constraint(role, constraint_values)
                else:
                    raise ValueError(f"Unsupported type for role value constraint: {extends._name}.")

    def _build_reading(self, model, reading_order):
        object_types = self._parser.object_types()
        rel_args = []
        if reading_order.front_text is not None:
            rel_args.append(f"{reading_order.front_text} ")
        for rdo_role in reading_order.roles:
            p = model.lookup_concept(object_types[rdo_role.role.player].name)
            rel_args.append(f"{{{p}}}") if rdo_role.role.name == "" else rel_args.append(f"{{{rdo_role.role.name}:{p}}}")
            if rdo_role.text is not None:
                rel_args.append(f" {rdo_role.text} ")
        return ''.join(rel_args).strip()

    def _rename_to_ref_mode(self, model, first_player_entity, fact_type):
        identifier_fact_type_to_entity_type = self._parser.identifier_fact_type_to_entity_type()
        entity_type = identifier_fact_type_to_entity_type.get(fact_type)
        if entity_type:
            player = model.lookup_concept(entity_type.name)
            if first_player_entity._id == player._id:
                return entity_type.ref_mode
        return None

    def _pick_short_name(self, model, player_entity, fact_type, reading):
        ref_mode = self._rename_to_ref_mode(model, player_entity, fact_type)
        if ref_mode is not None:
            rel_name = ref_mode.lower()
        else:
            rel_name = reading.rai_way_name()
        return rel_name

    def _role_id_to_role_object(self, model, role_obj):
        player_entity = model.lookup_concept(self._parser.object_types()[role_obj.player].name)
        rel = self._relationships[role_obj.relationship_name]
        if role_obj.name == "":
            idx = rel._field_names.index(player_entity._name.lower())
        else:
            idx = rel._field_names.index(role_obj.name)
        return rel[idx]

    def _get_roles_from_orm_constraint(self, model, orm_constraint):
        roles = self._parser.roles()
        # Role sequences
        if all(isinstance(item, list) for item in orm_constraint.roles):
            return [[self._role_id_to_role_object(model, roles[ro]) for ro in ro_list] for ro_list in orm_constraint.roles]
        # Role list
        else:
            return [self._role_id_to_role_object(model, roles[ro]) for ro in orm_constraint.roles]

    # A Value type that plays a given role is an Enum when there is a role value constraint containing only string
    # values declared on this ValueType, and it does not play a role in any other relationship
    def _is_value_type_playing_role_an_enum(self, role_id):
        role_value_constraint = self._parser.role_value_constraints().get(role_id)
        if role_value_constraint:
            vt_id = role_value_constraint.role.player
            return self._is_value_type_an_enum(vt_id) or (
                    role_value_constraint.values
                    and all(isinstance(v, str) for v in role_value_constraint.values)
                    and self._count_relationships_for_player(vt_id) == 1
            )
        return False

    # A Value Type is an Enum when it has a value type constraint containing only string values
    def _is_value_type_an_enum(self, vt_id):
        value_type_constraint = self._parser.value_type_value_constraints().get(vt_id)
        return (value_type_constraint and value_type_constraint.ranges
                and all(isinstance(v, str) for v in value_type_constraint.ranges))

    def _count_relationships_for_player(self, player_id):
        count = 0
        for v in self._parser._role_to_player.values():
            if v == player_id:
                count += 1
        return count

    @staticmethod
    def _get_alternative_identifying_reading(entity, id_relationship):
        if len(id_relationship._readings) > 0:
            for rd in id_relationship._readings[1:]:
                roles = rd._roles()
                if roles[0]._concept._name == entity._name:
                    return rd
        return None

    @staticmethod
    def _build_constraint_range(range: ORMRange, tp: type):
        if range.range_from == range.range_to:
            return tp(range.range_from)
        return Range.between(tp(range.range_from), tp(range.range_to))

    @staticmethod
    def _build_constraint_ranges(constraint, tp: type):
        constraint_ranges = []
        for rg in constraint.ranges:
            if isinstance(rg, ORMRange):
                constraint_ranges.append(ORMAdapterQB._build_constraint_range(rg, tp))
            else:
                constraint_ranges.append(tp(rg))
        return constraint_ranges

    @staticmethod
    def _set_role_verbalization(relationship, rdo):
        for idx, ro in enumerate(rdo.roles):
            qb_role = relationship._roles()[idx]
            text_chunks = [None, None]
            if ro.prefix:
                text_chunks[0] = ro.prefix
            if ro.postfix:
                text_chunks[1] = ro.postfix
            if text_chunks[0] or text_chunks[1]:
                qb_role.verbalization(*text_chunks)

    @staticmethod
    def _validate_role_verbalization(reading_orders):
        if len(reading_orders) > 1:
            relationship = reading_orders[0].roles[0].role.relationship_name
            for ro in reading_orders[0].roles:
                prefix = ro.prefix
                postfix = ro.postfix
                player = ro.role.player
                #  A role player must have the same prefix in all readings
                if not all(alt_ro.prefix == prefix for rdo in reading_orders[1:] for alt_ro in rdo.roles if alt_ro.role.player == player):
                    warnings.warn(f"Ignoring incompatible prefix in alternative readings of {relationship}.")
                #  A role player must have the same postfix in all readings
                if not all(alt_ro.postfix == postfix for rdo in reading_orders[1:] for alt_ro in rdo.roles if alt_ro.role.player == player):
                    warnings.warn(f"Ignoring incompatible postfix in alternative readings of {relationship}")



from relationalai.early_access.builder import Model, select, define, data, String
import pandas as pd

model = Model("movies", dry_run=False)
Concept, Relationship = model.Concept, model.Relationship

Movie = Concept("Movie")
Person = Concept("Person")
Directed = Concept("Directed")
ActedIn = Concept("ActedIn")


data(pd.read_csv("movies_data/movies.csv")).into(Movie, keys=["id"])
data(pd.read_csv("movies_data/person.csv")).into(Person, keys=["id"])
data(pd.read_csv("movies_data/directed.csv")).into(Directed, keys=["id"])
data(pd.read_csv("movies_data/actedin.csv")).into(ActedIn, keys=["id"])
# data(pd.read_csv("examples/builder/actors.csv")).into(Person, keys=["actor_id"])

# select(Person.name, Movie.title).where(
#     Directed.from_ == Person.id,
#     Directed.to == Movie.id
# ).inspect()

edge = Relationship("edge3 {String} {String} {String}")

define(edge(Person.name, Movie.title, "directed")).where(
    Directed.from_ == Person.id,
    Directed.to == Movie.id
)

define(edge(Person.name, Movie.title, "acted_in")).where(
    ActedIn.from_ == Person.id,
    ActedIn.to == Movie.id
)

a, b, label = String.ref(), String.ref(), String.ref()
select(a, b, label).where(edge(a, b, label)).inspect()




# # define author relationship by matching author_id
# define(Book.author(Author)).where(Author.author_id == Book.author_id)

# # get each book title with its author's name
# select(Book.title, Book.author.author_name).inspect()

# # get each pair of books published n the same year
# b = Book
# b1 = Book.ref()
# b2 = Book.ref()
# where(
#     b.publication_year == b1.publication_year,
#     b.title < b1.title
# ).select(b.title, b1.title).inspect()

# # get each triple of books published in the same year
# where(
#     b.publication_year == b1.publication_year,
#     b1.publication_year == b2.publication_year,
#     b.title < b1.title,
#     b1.title < b2.title
# ).select(b.title, b1.title, b2.title).inspect()

# # create a relationship for the author's published genres
# Author.genres = Relationship("{Author} has published {genres:str*}")
# # the Book.genre isn't necessary in the when, but the current rel
# # backend fails otherwise
# define(Author.genres(Book.genre)).where(Book.author == Author, Book.genre)

# select(Author.author_name, Author.genres).inspect()

# # get each author that has published in the dystopian genre and another genre
# select(Author.author_name).where(
#     Author.genres == "Dystopian",
#     Author.genres.ref() != "Dystopian"
# ).inspect()



from __future__ import annotations
from collections import defaultdict
from contextvars import ContextVar
import numbers
from typing import Any, Generic, Iterable, List, Set, TypeVar, cast
from typing_extensions import TypeGuard

from .rel_emitter import Emitter

from .metamodel import Behavior, Builtins, Action, ActionType, Type, Var, Task
from . import metamodel as m
from . import compiler as c
from .clients import config
from .dsl import build
from . import dsl

gather_vars = m.Utils.gather_vars
gather_task_vars = m.Utils.gather_task_vars

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def is_static(x:Any):
    if isinstance(x, m.Action):
        return all([is_static(z) for z in x.bindings.values()])
    if isinstance(x, Var):
        return x.value is not None
    if isinstance(x, Type):
        return True
    if isinstance(x, str):
        return True
    if isinstance(x, numbers.Number):
        return True
    if isinstance(x, list):
        return all(is_static(i) for i in x)
    if isinstance(x, tuple):
        return all(is_static(i) for i in x)
    if isinstance(x, dict):
        return all(is_static(i) for i in x.values())
    return False

def prepend_bindings(prefix:List[Var], action:Action):
    neue_bindings = {}
    for i, var in enumerate(prefix):
        neue_bindings[Builtins.Relation.properties[i]] = var
    prop_len = len(neue_bindings)
    for v in action.bindings.values():
        neue_bindings[Builtins.Relation.properties[prop_len]] = v
        prop_len += 1
    action.bindings = neue_bindings

def add_deps_to_binds(root:Task, bound_item:Task, deps:List[Var], replace_task=None):
    has_binds = False
    for item in root.items:
        if item.is_subtask_call():
            assert isinstance(item.entity.value, Task)
            has_binds |= add_deps_to_binds(item.entity.value, bound_item, deps, replace_task=replace_task)
        if item.action == ActionType.Bind and item.entity.value == bound_item:
            if replace_task:
                item.entity.value = replace_task
            prepend_bindings(deps, item)
            has_binds = True
    return has_binds

def flatten(items:List[Any]):
    neue = []
    for item in items:
        if isinstance(item, list):
            neue.extend(flatten(item))
        elif isinstance(item.value, list):
            for i in item.value:
                if isinstance(i, Var):
                    neue.append(i)
        elif isinstance(item, Var):
            neue.append(item)
    return neue

def is_base_type(val:m.Value|None) -> TypeGuard[m.Type]:
    return type(val) is m.Type and (not val.isa(Builtins.Relation) or val.isa(Builtins.Type))

def is_prop(val:m.Value|None) -> TypeGuard[m.Property]:
    return isinstance(val, m.Property) or isinstance(val, Type) and val.isa(Builtins.Property)

def pr(*args:Any):
    def stringify(x):
        if isinstance(x, list) or isinstance(x, set):
            return [stringify(i) for i in x]
        if isinstance(x, OrderedSet):
            return stringify(list(x))
        return str(x)
    return print(*[stringify(x) for x in args])

intermediate_annotations = ContextVar("intermediate_annotations", default=[Builtins.PipelineAnnotation])

def annotate_intermediate(task: Task):
    for annotation in intermediate_annotations.get():
        if annotation not in task.parents:
            task.parents.append(annotation)

#--------------------------------------------------
# OrderedSet
#--------------------------------------------------

T = TypeVar('T')
class OrderedSet(Generic[T]):
    def __init__(self):
        self.set:Set[T] = set()
        self.list:List[T] = []

    def add(self, item:T):
        if item not in self.set:
            self.set.add(item)
            self.list.append(item)

    def update(self, items:Iterable[T]):
        for item in items:
            self.add(item)

    def remove(self, item:T):
        if item in self.set:
            self.set.remove(item)
            self.list.remove(item)

    def clear(self):
        self.set.clear()
        self.list.clear()

    def __getitem__(self, ix):
        if len(self.list) <= ix:
            return None
        return self.list[ix]

    def __contains__(self, item:T):
        return item in self.set

    def __bool__(self):
        return bool(self.set)

    def __iter__(self):
        return iter(self.list)

    def __len__(self):
        return len(self.list)

    def __or__(self, other: OrderedSet[T]) -> OrderedSet[T]:
        result = OrderedSet[T]()
        result.update(self.list)
        result.update(other.list)
        return result

class ActionSet(OrderedSet[Action]):
    def __init__(self):
        super().__init__()

    def add(self, item:Action):
        for sub in self.list:
            if item.equiv(sub):
                return
        super().add(item)

#--------------------------------------------------
# Flow
#--------------------------------------------------

class FlowFrame:
    def __init__(self, flow:Flow, parent=None) -> None:
        self.flow = flow
        self.parent = parent
        self.sources = defaultdict(ActionSet)
        self.refs:defaultdict[Var, ActionSet] = defaultdict(ActionSet)
        self.queued:list[Action] = []
        self.multi = set()
        self.var_counts = defaultdict(int)
        self.var_max_order = defaultdict(int)

        if parent:
            for k, v in parent.sources.items():
                self.sources[k].update(v)
            for k, v in parent.refs.items():
                self.refs[k].update(v)
            for v in parent.multi:
                self.multi.add(v)
            for v, c in parent.var_counts.items():
                self.var_counts[v] = c
            for v, c in parent.var_max_order.items():
                self.var_max_order[v] = c

class Flow:
    multi_names = set()

    def __init__(self) -> None:
        self.tasks:list[Task] = []
        self.stack:list[FlowFrame] = []
        self.item_order = dict()
        self.item_is_source = defaultdict(bool)

    def push(self, no_parent=False):
        parent = None if no_parent else self.stack[-1] if self.stack else None
        self.stack.append(FlowFrame(self, parent))

    def pop(self):
        return self.stack.pop()

    @property
    def sources(self):
        return self.stack[-1].sources

    @property
    def refs(self):
        return self.stack[-1].refs

    @property
    def queued(self):
        return self.stack[-1].queued

    @property
    def multi(self):
        return self.stack[-1].multi

    def reset(self):
        self.tasks.clear()

    def peek_action(self, item:Action):
        if item not in self.item_order:
            self.item_order[item] = len(self.item_order)
        return self.item_order[item]

    def action(self, item:Action):
        self.peek_action(item)
        params = [p for p in flatten(item.params_list()) if p.value is None]
        if item.entity.isa(Builtins.GlobalFilter) or not params:
            self.queued.append(item)
            return

        has_source = False
        for param in params:
            if not self.sources[param]:
                has_source |= self.var_has_uses(param)
                self.sources[param].add(item)
            self.refs[param].add(item)

        self.item_is_source[item] = has_source

        if is_prop(item.entity.value):
            prop = item.entity.value
            var = params[-1]
            if var.value is None:
                self.sources[var].add(item)
            else:
                self.sources[params[0]].add(item)
            if prop.name in self.multi_names or Builtins.MultiValued in prop.parents:
                self.multi_names.add(prop.name)
                if var and var.value is None:
                    self.multi.add(var)
        elif is_base_type(item.entity.value) or item.entity.isa(Builtins.Filter) or item.entity.isa(Builtins.Extender) or not has_source:
            for param in params:
                self.sources[param].add(item)

    def remove_action(self, item:Action):
        for param in item.params_list():
            self.refs[param].remove(item)
            self.sources[param].remove(item)

    def action_has_source(self, item:Action):
        return self.item_is_source[item]

    def var(self, var:Var, max_order:int):
        self.stack[-1].var_counts[var] += 1
        self.stack[-1].var_max_order[var] = max(max_order, self.stack[-1].var_max_order.get(var, 0))

    def var_order(self, var:Var):
        return self.stack[-1].var_max_order.get(var, 0)

    def var_uses(self, var:Var):
        return self.stack[-1].var_counts[var] - 1

    def var_has_uses(self, var:Var):
        return self.stack[-1].var_counts[var] > 1

    def keys(self, var:Var):
        ks = []
        for ref in self.sources[var]:
            params = ref.params_list()
            if is_prop(ref.entity.value) and params[1] == var:
                ks.append(params[0])
            if is_base_type(ref.entity.value):
                return [var]
            if (ref.entity.isa(Builtins.Aggregate)
                and not ref.entity.isa(Builtins.Extender)
                and params[-1] == var
                and len(params) > 1):
                # if the source is an aggregate and there are grouping vars,
                # then the key is the grouping vars. Any vars before the agg value
                # are the group vars. Sorts (which are extender aggs) are ignored
                # as this would add the sort projection as keys.
                return params[:-1]
        return ks

    def _action_vars(self, item:Action, to_check:List[Var]):
        params = [v for v in flatten(item.params_list()) if v not in to_check and v.value is None]
        return params

    def demand_filters(self, var:Var, outs:ActionSet, to_check:List[Var], seen:set[Var]):
        if var in seen:
            return
        seen.add(var)
        if len(self.sources[var]) > 1:
            for sub in self.sources[var]:
                outs.add(sub)
                to_check.extend(self._action_vars(sub, to_check))
        for ref in self.refs[var]:
            if ref in outs:
                continue
            for param in self._action_vars(ref, to_check):
                self.demand_filters(param, outs, to_check, seen)

    def demand(self, var:Var, outs:ActionSet, to_check:List[Var], seen:set[Var]):
        sources = self.sources[var]
        seen = seen.copy()
        for ref in self.refs[var]:
            if ref in sources:
                outs.add(ref)
                to_check.extend(self._action_vars(ref, to_check))
            elif ref in outs:
                continue
            for param in self._action_vars(ref, to_check):
                self.demand_filters(param, outs, to_check, seen)

        return outs

    def demand_vars(self, vars:Iterable[Var], seen:set[Var]|None=None):
        outs:ActionSet = ActionSet()
        if seen is None:
            seen = set()
        to_check = list(vars)

        if self.queued:
            for item in self.queued:
                outs.add(item)
                to_check.extend(self._action_vars(item, to_check))
            self.queued.clear()

        for var in to_check:
            if var in seen:
                continue
            self.demand(var, outs, to_check, seen)
            seen.add(var)

        return sorted(outs.list, key=lambda x: self.item_order.get(x, 0))

    def task(self, task:Task):
        self.tasks.append(task)

    def find_uses(self, task:Task, uses:dict[Task, list[tuple[Task, Action]]]):
        for item in task.items:
            if item.entity.isa(Builtins.Intermediate) and item.action == ActionType.Get:
                assert isinstance(item.entity.value, Task)
                uses[item.entity.value].append((task, item))
            for v in item.params_list():
                if isinstance(v.value, Task) and v.value.items:
                    self.find_uses(v.value, uses)

    def finalize(self):
        assert not self.stack, "Flow still has frames on the stack"

        auto_inline = True
        if auto_inline:
            uses = defaultdict(list)
            replaced = {}
            collapsed_tasks = []
            for task in reversed(self.tasks):
                self.find_uses(task, uses)
                if task.isa(Builtins.Intermediate) and len(uses[task]) == 1:
                    sub, sub_action = uses[task][0]
                    if replaced.get(sub):
                        sub = replaced[sub]
                    replaced[task] = sub
                    sub.items.remove(sub_action)
                    for item in reversed(task.items):
                        if not (item.action == ActionType.Bind and item.entity.value is task):
                            sub.items.insert(0, item)
                    dedeuped = ActionSet()
                    for item in sub.items:
                        dedeuped.add(item)
                    sub.items = dedeuped.list
                else:
                    collapsed_tasks.append(task)

            items = [
                build.relation_action(ActionType.Call, task, []) for task in reversed(collapsed_tasks)
            ]
        else:
            items = [
                build.relation_action(ActionType.Call, task, []) for task in self.tasks
            ]

        return items


#--------------------------------------------------
# Dataflow
#--------------------------------------------------

class Dataflow(c.Pass):
    def __init__(self, copying=True) -> None:
        super().__init__(copying)
        self.flow = Flow()

    def reset(self):
        super().reset()
        self.flow.reset()

    #--------------------------------------------------
    # Query
    #--------------------------------------------------

    def query(self, query: Task, parent=None):
        self.flow.push()
        self.handle_items(query.items)
        self.flow.pop()
        final = self.flow.finalize()
        if final:
            query.behavior = Behavior.Sequence
        query.items = final

    def handle_items(self, items:list[Action]):
        ix = 0
        # count the number of times a variable appears in the items
        for item in items:
            item_ix = self.flow.peek_action(item)
            for v in item.vars(recursive=True, with_subs=True):
                self.flow.var(v, item_ix)

        while ix < len(items):
            item = items[ix]
            if item.is_subtask_call():
                task = item.entity.value
                assert isinstance(task, Task)
                if task.behavior == Behavior.Union:
                    self.handle_union(item)
                elif task.behavior == Behavior.OrderedChoice:
                    self.handle_choice(item)
                else:
                    self.handle_subtask(item)
                ix += 1

            # gather all the consecutive binds and handle them together
            elif item.action.is_effect():
                sub_ix = ix + 1
                while sub_ix < len(items) and items[sub_ix].action.is_effect():
                    sub_ix += 1
                binds = items[ix:sub_ix]
                self.handle_binds(binds)
                ix = sub_ix

            elif item.entity.isa(Builtins.Aggregate):
                self.handle_aggregate(item)
                ix += 1

            elif item.entity.isa(Builtins.Quantifier):
                self.handle_quantifier(item)
                ix += 1

            elif item.entity.isa(Builtins.Install):
                sub_ix = ix + 1
                while sub_ix < len(items) and items[sub_ix].entity.isa(Builtins.Install):
                    sub_ix += 1
                binds = items[ix:sub_ix]
                self.handle_installs(binds)
                ix = sub_ix

            # otherwise we consume the action and decide what to do with it later
            else:
                self.flow.action(item)
                ix += 1

    def determine_keys(self, var:Var, seen:set[Var]|None=None):
        if var in self.flow.multi:
            return [var]
        if keys := self.flow.keys(var):
            return keys

        if seen is None:
            seen = set()
        seen.add(var)
        keys = []
        for ref in self.flow.sources[var]:
            # ignore filters, they can't produce values
            if ref.entity.isa(Builtins.Filter):
                continue
            if ref.entity.isa(Builtins.Expensive):
                return [var]
            for v in flatten(ref.params_list()):
                if v not in seen:
                    keys.extend(self.determine_keys(v, seen))
        return keys

    def gather_incomputable_vars(self, var:Var, vars:OrderedSet[Var], seen:set[Var]):
        if var in seen or var.value is not None or not self.flow.var_has_uses(var):
            return
        seen.add(var)
        keys = self.flow.keys(var)
        is_multi = var in self.flow.multi
        if is_multi:
            if len(self.flow.sources[var]) > 1:
                vars.add(var)
            if len(self.flow.refs[var]) > 1:
                vars.add(var)
        elif keys:
            vars.update(keys)
            seen.update(keys)
        # else:
        sources = self.flow.sources[var]
        for ref in sources:
            if ref.entity.isa(Builtins.Expensive) and self.flow.var_uses(var) > 0:
                vars.add(var)
            elif ref.action == ActionType.Call \
                and not ref.entity.isa(Builtins.Filter) \
                and not ref.entity.isa(Builtins.SingleValued):
                vars.add(var)
            for v in flatten(ref.params_list()):
                self.gather_incomputable_vars(v, vars, seen)
        if sources:
            first = sources.list[0]
            vs = {
                param for param in flatten(first.params_list())
                if param.value is None \
                    and self.flow.sources[param][0] == first \
                    and self.flow.var_has_uses(param)
            }
            if len(vs) > 1:
                vars.update(vs)

    def gather_action_vars(self, actions:Iterable[Action]):
        vars = OrderedSet()
        seen = set()
        all_actions = list(actions)
        for action in all_actions:
            for v in action.params_list():
                self.gather_incomputable_vars(v, vars, seen)
            if action.is_subtask_call():
                assert isinstance(action.entity.value, Task)
                all_actions.extend(action.entity.value.items)
        return vars

    def _downstream(self, min_order:int, var:Var, seen:set[Var], available_vars:set[Var], depth=1):
        # This function finds the most downstream variable that is sourced from var and
        # is referenced at or after min_order
        if self.flow.var_order(var) >= min_order:
            return var

        seen.add(var)
        desired = OrderedSet()
        for ref in self.flow.refs[var]:
            for dep in ref.vars(recursive=True, with_subs=True):
                if dep.value is None and dep not in seen:
                    downstream = self._downstream(min_order, dep, seen, available_vars, depth+1)
                    if downstream:
                        desired.add(downstream)

        # if any desired variables aren't actually available then we must return this var
        # as it'll be used to compute those later
        if any([v not in available_vars for v in desired]):
            return var

        # if there are multiple desired, then this is the most upstream ancestor
        if len(desired) > 1:
            return var

        return desired[0]

    def create_intermediate(self, actions:list[Action]):
        vars = self.gather_action_vars(actions)

        min_order = min([self.flow.item_order.get(x, 0) for x in actions])
        outs = self.flow.demand_vars(vars)

        get = outs[0] if outs else None
        if len(outs) > 1:
            # find the minimal set of vars that are needed downstream and use those
            # as the head of the intermediate
            cur_vars = set(vars)
            out_vars = OrderedSet()
            if len(vars) > 1:
                available_vars = set(flatten([action.params_list() for action in outs]))
                for v in vars:
                    down = self._downstream(min_order, v, cur_vars.copy(), available_vars)
                    if down:
                        out_vars.add(down)
            else:
                out_vars = vars

            # this intermediate will now be the source for those
            for v in out_vars:
                self.flow.sources[v].clear()
                to_remove = [ref for ref in self.flow.refs[v] if ref in outs]
                for ref in to_remove:
                    self.flow.refs[v].remove(ref)

            intermediate = Task(items=[*outs], parents=[Builtins.Intermediate, Builtins.Expensive])
            intermediate.items.append(build.relation_action(ActionType.Bind, intermediate, out_vars))
            self.flow.task(intermediate)
            get = build.relation_action(ActionType.Get, intermediate, out_vars)
            self.flow.action(get)
            for action in outs:
                action_params = action.params_list()
                keys = set()
                can_remove = True
                for param in action_params:
                    keys.update(self.determine_keys(param))
                    if self.flow.sources[param][0] == action:
                        can_remove = False
                        break
                if keys.issubset(vars) and can_remove:
                    self.flow.remove_action(action)

            vars = list(out_vars)
        return get, vars

    def _split_return(self, bind, params, vars, ret_vars, get, demand=True):
        action_type = bind.action
        is_export = bind.entity.isa(Builtins.ExportReturn)
        for (ix, v) in enumerate(params):
            sub = list(self.flow.demand_vars([v], seen=set(vars))) if demand else []
            if get:
                sub.insert(0, get)
            col = Var(Builtins.Symbol, value=f"col{ix:03}")
            if is_export:
                rel = dsl.rel.Export_Relation._rel
                uuid_str = dsl.rel.uuid_string._rel
                v2 = Var()
                sub.append(build.relation_action(ActionType.Call, dsl.rel.pyrel_default._rel, [uuid_str, v, v, v2]))
                sub.append(build.relation_action(action_type, rel, [col, *ret_vars, v2]))
            else:
                rel = dsl.rel.output.cols._rel
                sub.append(build.relation_action(action_type, rel, [col, *ret_vars, v]))
            self.flow.task(Task(items=sub))


    def handle_binds(self, binds:List[Action]):
        get, vars = self.create_intermediate(binds)

        # for each bind, generate the output action
        statics = []
        for bind in binds:
            self.flow.push()
            action_type = bind.action
            params = bind.params_list()
            if is_static(bind) and not bind.entity.isa(Builtins.Return):
                statics.append(bind)
            elif (bind.entity.isa(Builtins.Distinct) and not set(vars).issubset(set(params))) or (bind.entity.isa(Builtins.Return) and set(vars) == set(params)):
                sub_raw = list(self.flow.demand_vars(params, seen=set(vars)))
                sub = []
                for item in sub_raw:
                    if isinstance(item.entity.value, m.Property):
                        [k, v] = item.params_list()
                        sub.append(build.relation_action(ActionType.Call, dsl.rel.pyrel_default._rel, [item.entity.value, dsl.rel.Missing, k, v]))
                    else:
                        sub.append(item)
                if get and get not in sub:
                    sub.insert(0, get)

                intermediate = Task(items=sub)
                hash_var = Var(name="hash")
                sub.append(build.call(Builtins.make_identity, [Var(value=list(params)), hash_var]))
                sub.append(build.relation_action(action_type, intermediate, params + [hash_var]))
                self.flow.task(Task(items=sub))
                intermediate_get = build.relation_action(ActionType.Get, intermediate, params + [hash_var])
                self._split_return(bind, params, vars, [hash_var], intermediate_get, demand=False)

            elif bind.entity.isa(Builtins.Return):
                ret_vars = self.gather_action_vars([bind])
                if not ret_vars:
                    ret_vars = [Var(value=bind.id)]
                self._split_return(bind, params, vars, ret_vars, get)
            else:
                sub = list(self.flow.demand_vars(params, seen=set(vars)))
                if get and get not in sub:
                    sub.insert(0, get)
                sub.append(bind)
                self.flow.task(Task(items=sub))
            self.flow.pop()

        if statics:
            self.flow.push()
            if get:
                statics.insert(0, get)
            self.flow.task(Task(items=statics))
            self.flow.pop()

    def handle_aggregate(self, item:Action):
        agg = cast(Task, item.entity.value)
        is_extender = agg.isa(Builtins.Extender)
        (args, group, pre_args, ret) = item.params_list()
        group_vars = cast(List[Var], group.value)
        arg_vars = cast(List[Var], args.value)
        pre_arg_vars = cast(List[Var], pre_args.value)

        flattened_call = build.relation_action(ActionType.Get, agg, [*pre_arg_vars, *group_vars, *arg_vars])
        self.flow.item_order[flattened_call] = self.flow.item_order.get(item, 0)
        get, vars = self.create_intermediate([flattened_call])

        # to prevent shadowing errors we need to map the inner vars to new vars
        mapped = [Var(name=var.name, type=var.type, value=var.value) for var in arg_vars]

        # create the inner relation we'll aggregate over
        inner_items = list(self.flow.demand_vars([*arg_vars, *group_vars], seen=set(vars)))
        inner_vars = arg_vars.copy()
        if get:
            inner_items.insert(0, get)

        #de-shadow any arguments that are also in the group by
        for ix, arg in enumerate(arg_vars):
            if arg in group_vars:
                inner_vars[ix] = Var(name=f"inner_{arg.name}", type=arg.type, value=arg.value)
                inner_items.append(build.eq(inner_vars[ix], arg))

        inner = Task(items=inner_items)
        inner_items.append(build.relation_action(ActionType.Bind, inner, inner_vars))

        agg_bindings = [ret] if not is_extender else [ret, *mapped]
        agg_call = build.relation_action(ActionType.Call, agg, [*pre_arg_vars, Var(value=inner), *agg_bindings])


        outer = Task(items=[agg_call], parents=[Builtins.Aggregate])
        if is_extender:
            outer.parents.append(Builtins.Extender)
        outer_params = [*group_vars, ret] if not is_extender else [ret, *group_vars, *mapped]
        outer.items.append(build.relation_action(ActionType.Bind, outer, outer_params))
        self.flow.task(outer)

        get_params = outer_params if not is_extender else [ret, *group_vars, *arg_vars]
        get = build.relation_action(ActionType.Get, outer, get_params)
        self.flow.action(get)

    def handle_subtask(self, item:Action):
        subtask = cast(Task, item.entity.value)
        get, vars = self.create_intermediate(subtask.items)
        sub_items = [get] + subtask.items if get else subtask.items
        self.flow.push()
        self.handle_items(sub_items)
        self.flow.pop()

    def handle_quantifier(self, item:Action):
        quantifier = item.entity.value
        assert isinstance(quantifier, Task)
        group, task_var = [*item.bindings.values()]
        sub_task = task_var.value
        assert isinstance(sub_task, Task)

        if isinstance(group.value, list) and len(group.value):
            raise Exception("TODO: grouped quantifiers")

        shared = self.gather_action_vars(sub_task.items)
        if len(sub_task.items) > 1:
            task_to_quantify = Task(items=[
                *sub_task.items,
            ], parents=[Builtins.InlineAnnotation])
            task_to_quantify.items.append(build.relation_action(ActionType.Bind, task_to_quantify, shared))
            self.flow.task(task_to_quantify)
            action = build.relation_action(ActionType.Get, task_to_quantify, shared)
        else:
            cur = sub_task.items[0]
            bindings = {k: v if v in shared else Var(name="_") for k,v in cur.bindings.items()}
            action = Action(action=cur.action, entity=cur.entity, types=cur.types, bindings=bindings)


        # create the quantified task, which just gets the subtask
        quantifed_task = Task()
        quantifed_task.items.append(action)

        call = build.call(quantifier, [group, Var(value=quantifed_task)])
        self.flow.action(call)

    def handle_union(self, action:Action):
        union_task = action.entity.value
        assert isinstance(union_task, Task)
        union_task.parents.append(Builtins.Intermediate)

        self.flow.push()
        get, shared = self.create_intermediate(union_task.items)
        shared = list(shared)
        has_outputs = False
        for item in union_task.items:
            assert isinstance(item.entity.value, Task)
            sub = item.entity.value
            has_outputs |= add_deps_to_binds(sub, union_task, shared)
            self.flow.push()
            items = [get] + sub.items if get else sub.items
            self.handle_items(items)
            self.flow.pop()

        self.flow.pop()

        # Get the result of the union
        if has_outputs:
            get = build.relation_action(ActionType.Get, union_task, shared + action.params_list())
            # self.action(get)
            for v in shared:
                self.flow.sources[v].clear()
                self.flow.sources[v].add(get)
                # self.flow.refs[v].clear()
            self.flow.action(get)
            # for var in get.params_list():
            #     # self.flow.sources[var].append(get)
            #     self.flow.sources[var].add(get)

    def handle_choice(self, action:Action):
        choice = action.entity.value
        assert isinstance(choice, Task)
        choice.parents.append(Builtins.EDB)
        choice_params = action.params_list()
        prevs = []

        self.flow.push()
        get, shared = self.create_intermediate(choice.items)
        shared = list(shared)
        has_outputs = False

        for item in choice.items:
            cur_task = item.entity.value
            assert isinstance(cur_task, Task)
            # cur_task.parents.append(Builtins.InlineAnnotation)
            # add deps to the binds and also replace the task so that we write to the current
            # branch rather than the final return
            cur_has_outputs = add_deps_to_binds(cur_task, choice, shared, replace_task=cur_task)
            has_outputs |= cur_has_outputs

            # Negate all the previous branches to ensure we only return a value if
            # we're at our position in the order
            for prev in reversed(prevs):
                fetch = build.relation_action(ActionType.Get, prev, shared + [Var(name="_") for _ in choice_params])
                prev_task = Task(items=[fetch])
                cur_task.items.insert(0, build.call(Builtins.Not, [Var(value=[]), Var(value=prev_task)]))

            self.flow.push()
            if get:
                self.flow.queued.append(get)
            items = cur_task.items
            if not cur_has_outputs:
                items.append(build.relation_action(ActionType.Bind, cur_task, shared + choice_params))
            # items.append(
            #     build.relation_action(ActionType.Bind, choice, shared + choice_params),
            # )
            self.handle_items(items)
            self.flow.pop()

            self.flow.push(no_parent=True)
            self.handle_items([
                build.relation_action(ActionType.Get, cur_task, shared + choice_params),
                build.relation_action(ActionType.Bind, choice, shared + choice_params),
            ])
            self.flow.pop()

            prevs.append(cur_task)

        self.flow.pop()

        # Get the result of the choice
        if has_outputs:
            get = build.relation_action(ActionType.Get, choice, shared + choice_params)
            for v in shared:
                self.flow.sources[v].clear()
                self.flow.sources[v].add(get)
            self.flow.action(get)

    def handle_installs(self, installs:list[Action]):
        self.flow.task(Task(items=installs))


#--------------------------------------------------
# Shredder
#--------------------------------------------------

class Shredder(c.Pass):
    def query(self, query: Task, parent=None):
        neue_actions = []
        for item in query.items:
            if item.action not in [ActionType.Call, ActionType.Construct] and not item.entity.isa(Builtins.Relation):
                ident, action = item.entity, item.action
                for type in item.types:
                    neue_actions.append(build.relation_action(action, type, [ident]))
                for prop, value in item.bindings.items():
                    neue_actions.append(build.relation_action(action, prop, [ident, value]))
            else:
                walked = self.walk(item)
                neue_actions.append(walked)
        query.items = neue_actions

#--------------------------------------------------
# Splinter
#--------------------------------------------------

class Splinter(c.Pass):

    def query(self, query: Task, parent=None):
        grouped_effects = defaultdict(list)
        prev_fetches = []
        effects = []
        non_effects = []
        neue_items = []
        non_effects_vars = OrderedSet()

        def process_grouped_effects():
            nonlocal neue_items, grouped_effects, non_effects, effects, non_effects_vars
            if len(grouped_effects) > 1:
                non_effects_vars.update(gather_vars(non_effects))
                fetch = None
                if non_effects:
                    fetch = self.create_fetch(prev_fetches + non_effects, non_effects_vars)
                    neue_items.append(fetch)
                    assert isinstance(fetch.entity.value, Type)
                    prev_fetches.append(build.relation_action(ActionType.Get, fetch.entity.value, non_effects_vars))

                for (k, b) in grouped_effects.items():
                    effect_query = self.create_effect_query(b, non_effects_vars, prev_fetches)
                    neue_items.append(effect_query)
            elif grouped_effects:
                neue_items.append(build.call(Task(items=prev_fetches + non_effects + effects), []))

            grouped_effects.clear()
            non_effects.clear()
            effects.clear()

        for item in query.items:
            if item.action.is_effect():
                grouped_effects[(item.action, item.entity.value)].append(item)
                effects.append(item)
            else:
                if grouped_effects:
                    process_grouped_effects()
                non_effects.append(item)

        if grouped_effects:
            process_grouped_effects()
        elif non_effects:
            neue_items.append(build.call(Task(items=[*prev_fetches, *non_effects]), []))

        if len(neue_items) > 1:
            query.behavior = Behavior.Sequence
            query.items = neue_items

    #--------------------------------------------------
    # Subtask creation
    #--------------------------------------------------

    def create_fetch(self, non_effects: List[Action], effects_vars: Iterable[Var]):
        fetch = Task()
        annotate_intermediate(fetch)
        non_effects = non_effects.copy()
        non_effects.append(build.relation_action(ActionType.Bind, fetch, effects_vars))
        fetch.items = non_effects
        return build.call(fetch, [])

    def create_effect_query(self, effects: List[Action], effects_vars: Iterable[Var], fetches: Any):
        neue = Task()
        effects = effects.copy()
        for fetch in reversed(fetches):
            effects.insert(0, fetch)
        neue.items = effects
        return build.call(neue, [])

#--------------------------------------------------
# SetCollector
#--------------------------------------------------

set_types = [ActionType.Bind, ActionType.Persist, ActionType.Unpersist]

class SetCollector(c.Pass):
    def query(self, query: Task, parent=None):
        binds = [i for i in query.items if i.action in set_types]
        if len(binds) > 1:
            neue_items = []
            for item in query.items:
                if item.action not in set_types:
                    neue_items.append(item)
            neue_items.extend(self.create_raw(binds))
            query.items = neue_items

    def create_raw(self, binds: List[Action]):
        vals = [Var(value=[]) for i in range(len(binds[0].bindings))]
        vars = [Var() for v in vals]

        for bind in binds:
            for ix, var in enumerate(bind.bindings.values()):
                cast(List[Var], vals[ix].value).append(var)

        return [
            build.relation_action(ActionType.Get, Builtins.RawData, vals + vars),
            build.relation_action(binds[0].action, cast(Type, binds[0].entity.value), vars)
        ]

#--------------------------------------------------
# Compiler
#--------------------------------------------------

class Clone(c.Pass):
    pass

class Compiler(c.Compiler):
    def __init__(self, config:config.Config):
        self.config = config
        super().__init__(Emitter(config), [
            Clone(),
            Shredder(),
            Dataflow(),
            Splinter(),
            SetCollector(),
        ])

    def compile(self, task: Task):
        token = None
        if self.config.get("use_inlined_intermediates", False):
            token = intermediate_annotations.set([Builtins.InlineAnnotation])

        try:
            return super().compile(task)
        finally:
            if token:
                intermediate_annotations.reset(token)



from relationalai.early_access.dsl.core.stack import Stack


class ContextStack(Stack):

    def __init__(self):
        super().__init__()

    def root_context(self):
        ctx = self.top()
        if ctx is None:
            raise Exception('Context stack is empty')
        return ctx


# For builder components.
from relationalai.early_access.builder import Model
from relationalai.early_access.builder import Integer
from relationalai.early_access.builder import define, max, sum, select, not_

from relationalai.early_access.paths.graph import Graph
from relationalai.early_access.paths.path_algorithms.one_sided_ball_upto import ball_upto
from relationalai.early_access.paths.path_algorithms.one_sided_ball_repetition import ball_with_repetition


def compute_usp(g: Graph, Source, Target, max_length = None):
    # Computes the Union of Shortest Paths (USP) from a source set to a destination set
    # USP is a subgraph of the original graph

    edge = g.Edge
    Node = g.Node

    n, d = Integer.ref(), Integer.ref()
    tgt, u, v = Node.ref(), g.Node.ref(), g.Node.ref()

    if max_length is None:
        ball = ball_upto(g, Source, Target)
        usp_nodes = g.model.Relationship(f"usp_nodes1 {{{Node}}}")
        usp = g.model.Relationship(f"usp1 {{{Node}}} {{{Node}}}")
        boundary = g.model.Relationship(f"boundary1 {{{Node}}}")

        # propagate backwards to find the nodes that are in the USP:
        define(boundary(tgt)).where(
            Target(tgt),
            ball(n, tgt)
        )

        define(usp_nodes(tgt)).where(
            boundary(tgt)
        )

        define(usp_nodes(u)).where(
            usp(u, v)
        )

        define(usp(u, v)).where(
            usp_nodes(v),
            ball(d, v),
            ball(d - 1, u),
            edge(u, v)
        )
    else:
        ball = ball_with_repetition(g, Source, max_length)
        max_dist = g.model.Relationship("max_dist1 {Integer}")
        NodeDist = g.model.Concept("NodeDist")
        usp = g.model.Relationship(f"usp2 {{{NodeDist}}} {{{NodeDist}}}")
        boundary = g.model.Relationship(f"boundary2 {{{NodeDist}}}")

        define(NodeDist.new(node = u, layer = n)).where(
            ball(n, u)
        )

        define(max_dist(max(n))).where(
            ball(n, tgt),
            Target(tgt)
        )

        uu, vv, ww = NodeDist.ref(), NodeDist.ref(), NodeDist.ref()

        define(boundary(uu)).where(
            NodeDist(uu),
            max_dist(n),
            Target(tgt),
            ball(n, tgt),
            uu.node == tgt,
            uu.layer == n
        )

        define(usp(uu, vv)).where(
            boundary(vv),
            NodeDist(uu),
            vv.node == v,
            vv.layer == n + 1,
            uu.node == u,
            uu.layer == n,
            ball(n, u),
            edge(u, v)
        )

        define(usp(uu, vv)).where(
            usp(vv, ww),
            vv.node == v,
            vv.layer == n + 1,
            uu.node == u,
            uu.layer == n,
            ball(n, u),
            edge(u, v)
        )


    return usp, boundary




def compute_nsp_from_usp(g: Graph, usp, Source, Target, Boundary):
    Node = g.Node

    n, m = Integer.ref(), Integer.ref()
    tgt, u, v = Node.ref(), Node.ref(), Node.ref()

    """if max_dist is None:
        nsp = g.model.Relationship(f"nsp2 {{{Node}}} {{Integer}}")
        define(nsp(tgt, 1)).where(
            Target(tgt),
            usp(u, tgt)
        )

        define(nsp(tgt, 1)).where(
            Target(tgt),
            Source(tgt)
        )

        define(nsp(u, sum(v, n).per(u))).where(
            nsp(v, n),
            usp(u, v)
        )
    else:
    """
    nsp = g.model.Relationship(f"nsp1 {{{Node}}} {{Integer}} {{Integer}}")

    define(nsp(tgt, n, 1)).where(
        Boundary(tgt, n)
    )

    define(nsp(u, n, sum(v, n + 1, m).per(u, n))).where(
        nsp(v, n + 1, m),
        usp(u, n, v),
        not_(Target(u))
    )

    define(nsp(u, n, 1 + sum(v, n + 1, m).per(u, n))).where(
        nsp(v, n + 1, m),
        usp(u, n, v),
        Target(u),
        not_(Boundary(u, n))
    )

    return nsp


def compute_nsp(g: Graph, Source, Target, max_length = None):
    # Computes the number of shortest paths (NSP) from a source set to a destination set

    if max_length is None:
        usp, Boundary = compute_usp(g, Source, Target)
        nsp = compute_nsp_from_usp(g, usp, Source, Target, Boundary)
    else:
        usp, Boundary  = compute_usp(g, Source, Target, max_length)
        nsp = compute_nsp_from_usp(g, usp, Source, Target, Boundary)

    return nsp


model = Model("test_usp_nsp_single", dry_run=False)

grid1 = Graph.construct_grid(model, 2, "undirected")

source1 = grid1.Node.new(row = 1, col = 1)
target1 = grid1.Node.new(row = 2, col = 2)

Source1 = model.Concept("Source1", extends=[grid1.Node])
Target1 = model.Concept("Target1", extends=[grid1.Node])

define(Source1(source1))
define(Target1(target1))
define(Target1(source1))

usp, boundary = compute_usp(grid1, Source1, Target1, 4)

usp.inspect()
#nsp = compute_nsp(grid1, Source1, Target1, 4)


u, v = grid1.Node.ref(), grid1.Node.ref(),
n, m = Integer.ref(), Integer.ref()

select(u, v).where(usp(u, v)).inspect()


# select(u.row, u.col, n, m).where(nsp(u, n, m)).inspect()


#pyright: reportPrivateImportUsage=false
from __future__ import annotations
import io
import os
from typing import Callable, Sequence, TextIO, cast, Any, List
from pathlib import Path
from InquirerPy.base.complex import FakeDocument
from prompt_toolkit.key_binding import KeyPressEvent
from prompt_toolkit.validation import ValidationError
import rich
from rich.console import Console
from rich.color import Color
import sys
from InquirerPy import inquirer, utils as inquirer_utils
from InquirerPy.base.control import Choice
import time
import threading
import itertools
import shutil
from wcwidth import wcwidth

from relationalai import debugging

from ..environments import runtime_env, NotebookRuntimeEnvironment, SnowbookEnvironment, HexEnvironment, JupyterEnvironment

#--------------------------------------------------
# Constants
#--------------------------------------------------

REFETCH = "[REFETCH LIST]"
MANUAL_ENTRY = "[MANUAL ENTRY]"

#--------------------------------------------------
# Style
#--------------------------------------------------

STYLE = inquirer_utils.get_style({
    "fuzzy_prompt": "#e5c07b"
}, False)

#--------------------------------------------------
# Helpers
#--------------------------------------------------

def rich_str(string:str, style:str|None = None) -> str:
    output = io.StringIO()
    console = Console(file=output, force_terminal=True)
    console.print(string, style=style)
    return output.getvalue()

def nat_path(path: Path, base: Path):
    resolved_path = path.resolve()
    resolved_base = base.resolve()
    if resolved_base in resolved_path.parents or resolved_path == resolved_base:
        return resolved_path.relative_to(resolved_base)
    else:
        return resolved_path.absolute()

def get_default(value:str|None, list_of_values:Sequence[str]):
    if value is None:
        return None
    list_of_values_lower = [v.lower() for v in list_of_values]
    value_lower = value.lower()
    if value_lower in list_of_values_lower:
        return value

#--------------------------------------------------
# Dividers
#--------------------------------------------------

def divider(console=None, flush=False):
    div = "\n[dim]---------------------------------------------------\n "
    if console is None:
        rich.print(div)
    else:
        console.print(div)
    if flush:
        sys.stdout.flush()

def abort():
    rich.print()
    rich.print("[yellow]Aborted")
    divider()
    sys.exit(1)

#--------------------------------------------------
# Prompts
#--------------------------------------------------

default_bindings = cast(Any, {
    "interrupt": [
        {"key": "escape"},
        {"key": "c-c"},
        {"key": "c-d"}
    ],
    "skip": [
        {"key": "c-s"}
    ]
})

def prompt(message:str, value:str|None, newline=False, validator:Callable|None = None, invalid_message:str|None = None) -> str:
    if value:
        return value
    if invalid_message is None:
        invalid_message = "Invalid input"
    try:
        result:str = inquirer.text(
            message,
            validate=validator,
            invalid_message=invalid_message,
            keybindings=default_bindings,
        ).execute()
    except KeyboardInterrupt:
        abort()
        raise Exception("Unreachable")
    if newline:
        rich.print("")
    return result

def select(message:str, choices:List[str|Choice], value:str|None, newline=False, **kwargs) -> str|Any:
    if value:
        return value
    try:
        result:str = inquirer.select(message, choices, keybindings=default_bindings, **kwargs).execute()
    except KeyboardInterrupt:
        abort()
        raise Exception("Unreachable")
    if newline:
        rich.print("")
    return result

def _enumerate_static_choices(choices: inquirer_utils.InquirerPyChoice) -> inquirer_utils.InquirerPyChoice:
    return [{"name": f"{i+1} {choice}", "value": choice} for i, choice in enumerate(choices)]

def _enumerate_choices(choices: inquirer_utils.InquirerPyListChoices) -> inquirer_utils.InquirerPyListChoices:
    if callable(choices):
        return lambda session: _enumerate_static_choices(choices(session))
    else:
        return _enumerate_static_choices(choices)

def _fuzzy(message:str, choices:inquirer_utils.InquirerPyListChoices, default:str|None = None, multiselect=False, show_index=False, **kwargs) -> str|list[str]:
    if show_index:
        choices = _enumerate_choices(choices)    

    try:
        kwargs["keybindings"] = default_bindings
        if multiselect:
            kwargs["keybindings"] = { # pylint: disable=assignment-from-no-return
                "toggle": [
                    {"key": "tab"},   # toggle choices
                ],
                "toggle-down": [
                    {"key": "tab", "filter":False},
                ],
            }.update(default_bindings)
            kwargs["multiselect"] = True

        # NOTE: Using the builtin `default` kwarg to do this also filters
        #       results which is undesirable and confusing for pre-filled
        #       fields, so we move the cursor ourselves using the internals :(
        prompt = inquirer.fuzzy(message, choices=choices, max_height=8, border=True, style=STYLE, **kwargs)
        prompt._content_control._get_choices(prompt._content_control.choices, default)

        return prompt.execute()
    except KeyboardInterrupt:
        return abort()

def fuzzy(message:str, choices:inquirer_utils.InquirerPyListChoices, default:str|None = None, show_index=False, **kwargs) -> str:
    return cast(str, _fuzzy(message, choices, default=default, show_index=show_index, **kwargs))

def fuzzy_multiselect(message:str, choices:inquirer_utils.InquirerPyListChoices, default:str|None = None, show_index=False, **kwargs) -> list[str]:
    return cast(list[str], _fuzzy(message, choices, default=default, show_index=show_index, multiselect=True, **kwargs))

def fuzzy_with_refetch(prompt: str, type: str, fn: Callable, *args, **kwargs):
    exception = None
    auto_select = kwargs.get("auto_select", None)
    not_found_message = kwargs.get("not_found_message", None)
    manual_entry = kwargs.get("manual_entry", None)
    items = []
    with Spinner(f"Fetching {type}", f"Fetched {type}"):
        try:
            items = fn(*args)
        except Exception as e:
            exception = e
    if exception is not None:
        rich.print(f"\n[red]Error fetching {type}: {exception}\n")
        return exception
    if len(items) == 0:
        if not_found_message:
            rich.print(f"\n[yellow]{not_found_message}\n")
        else:
            rich.print(f"\n[yellow]No valid {type} found\n")
        return None

    if auto_select and len(items) == 1 and items[0].lower() == auto_select.lower():
        return auto_select

    if manual_entry:
        items.insert(0, MANUAL_ENTRY)
    items.insert(0, REFETCH)

    passed_default = kwargs.get("default", None)
    passed_mandatory = kwargs.get("mandatory", False)

    rich.print("")
    result = fuzzy(
        prompt,
        items,
        default=get_default(passed_default, items),
        mandatory=passed_mandatory
    )
    rich.print("")

    while result == REFETCH:
        result = fuzzy_with_refetch(prompt, type, fn, *args, **kwargs)
    return result

def confirm(message:str, default:bool = False) -> bool:
    try:
        return inquirer.confirm(message, default=default, keybindings=default_bindings).execute()
    except KeyboardInterrupt:
        return abort()

def text(message:str, default:str|None = None, validator:Callable|None = None, invalid_message:str|None = None, **kwargs) -> str:
    if not invalid_message:
        invalid_message = "Invalid input"
    try:
        return inquirer.text(
            message,
            default=default or "",
            keybindings=default_bindings,
            validate=validator,
            invalid_message=invalid_message,
            **kwargs
        ).execute()
    except KeyboardInterrupt:
        return abort()

def password(message:str, default:str|None = None, validator:Callable|None = None, invalid_message:str|None = None) -> str:
    if invalid_message is None:
        invalid_message = "Invalid input"
    try:
        return inquirer.secret(
            message,
            default=default or "",
            keybindings=default_bindings,
            validate=validator,
            invalid_message=invalid_message
        ).execute()
    except KeyboardInterrupt:
        return abort()

def file(message: str, start_path:Path|None = None, allow_freeform=False, **kwargs) -> str|None:
    try:
        return FuzzyFile(message, start_path, allow_freeform=allow_freeform, max_height=8, border=True, style=STYLE, **kwargs).execute()
    except KeyboardInterrupt:
        return abort()

class FuzzyFile(inquirer.fuzzy):
    def __init__(self, message: str, initial_path: Path|None = None, allow_freeform = False,  *args, **kwargs):
        self.initial_path = initial_path or Path()
        self.current_path = Path(self.initial_path)
        self.allow_freeform = allow_freeform

        kwargs["keybindings"] = {
            **default_bindings,
            "answer": [
                {"key": os.sep},
                {"key": "enter"},
                {"key": "tab"},
                {"key": "right"}
            ],
            **kwargs.get("keybindings", {})
        }

        super().__init__(message, *args, **kwargs, choices=self._get_choices)

    def _get_prompt_message(self) -> List[tuple[str, str]]:
        pre_answer = ("class:instruction", f" {self.instruction} " if self.instruction else " ")
        result = str(nat_path(self.current_path, self.initial_path))

        if result:
            sep = " " if self._amark else ""
            return [
                ("class:answermark", self._amark),
                ("class:answered_question", f"{sep}{self._message} "),
                ("class:answer", f"{result}{os.sep if not self.status['answered'] else ''}"),
            ]
        else:
            sep = " " if self._qmark else ""
            return [
                ("class:answermark", self._amark),
                ("class:questionmark", self._qmark),
                ("class:question", f"{sep}{self._message}"),
                pre_answer
            ]

    def _handle_enter(self, event: KeyPressEvent) -> None:
        try:
            fake_document = FakeDocument(self.result_value)
            self._validator.validate(fake_document)  # type: ignore
            cc = self.content_control
            if self._multiselect:
                self.status["answered"] = True
                if not self.selected_choices:
                    self.status["result"] = [cc.selection["name"]]
                    event.app.exit(result=[cc.selection["value"]])
                else:
                    self.status["result"] = self.result_name
                    event.app.exit(result=self.result_value)
            else:
                res_value = cc.selection["value"]
                self.current_path /= res_value
                if self.current_path.is_dir():
                    self._update_choices()
                else:
                    self.status["answered"] = True
                    self.status["result"] = cc.selection["name"]
                    event.app.exit(result=str(nat_path(self.current_path, self.initial_path)))
        except ValidationError as e:
            self._set_error(str(e))
        except IndexError:
            self.status["answered"] = True
            res = self._get_current_text() if self.allow_freeform else None
            if self._multiselect:
                res = [res] if res is not None else []
            self.status["result"] = res
            event.app.exit(result=res)

    def _get_choices(self, _ = None):
        choices = os.listdir(self.current_path)
        choices.append("..")
        return choices

    def _update_choices(self):
        raw_choices = self._get_choices()
        cc = self.content_control
        cc.selected_choice_index = 0
        cc._raw_choices = raw_choices
        cc.choices = cc._get_choices(raw_choices, None)
        cc._safety_check()
        cc._format_choices()
        self._buffer.reset()

#--------------------------------------------------
# Spinner
#--------------------------------------------------

class Spinner:
    """Shows a spinner control while a task is running.
    The finished_message will not be printed if there was an exception and the failed_message is provided.
    """
    busy = False

    def __init__(
        self,
        message="",
        finished_message: str = "",
        failed_message=None,
        delay=None,
        leading_newline=False,
        trailing_newline=False,
    ):
        self.message = message
        self.finished_message = finished_message
        self.failed_message = failed_message
        self.spinner_generator = itertools.cycle(["▰▱▱▱", "▰▰▱▱", "▰▰▰▱", "▰▰▰▰", "▱▰▰▰", "▱▱▰▰", "▱▱▱▰", "▱▱▱▱"])
        self.is_snowflake_notebook = isinstance(runtime_env, SnowbookEnvironment)
        self.is_hex = isinstance(runtime_env, HexEnvironment)
        self.is_jupyter = isinstance(runtime_env, JupyterEnvironment)
        self.in_notebook = isinstance(runtime_env, NotebookRuntimeEnvironment)
        self.is_tty = sys.stdout.isatty()

        self._set_delay(delay)
        self.leading_newline = leading_newline
        self.trailing_newline = trailing_newline
        self.last_message = ""
        self.display = None
        # Add lock to prevent race conditions between spinner thread and main thread
        self._update_lock = threading.Lock()

    def _set_delay(self, delay: float|int|None) -> None:
        """Set appropriate delay based on environment and user input."""
        # If delay value is provided, validate and use it
        if delay:
            if isinstance(delay, (int, float)) and delay > 0:
                self.delay = float(delay)
                return
            else:
                raise ValueError(f"Invalid delay value: {delay}")
        # Otherwise, set delay based on environment
        elif self.is_hex:
            self.delay = 0 # Hex tries to append a new block each frame
        elif self.is_snowflake_notebook:
                self.delay = 0.5 # SF notebooks get bogged down
        elif self.in_notebook or self.is_tty:
            # Fast refresh for other notebooks or terminals with good printing support
            self.delay = 0.1
        else:
            # Otherwise disable the spinner animation entirely
            # for non-interactive environments.
            self.delay = 0

    def get_message(self, starting=False):
        max_width = shutil.get_terminal_size().columns
        spinner = "⏳⏳⏳⏳" if not self.is_tty and starting else next(self.spinner_generator)
        full_message = f"{spinner} {self.message}"
        if len(full_message) > max_width:
            return full_message[:max_width - 3] + "..."
        else:
            return full_message

    def update(self, message:str|None=None, color:str|None=None, file:TextIO|None=None, starting=False):
        # Use lock to prevent race conditions between spinner thread and main thread
        with self._update_lock:
            if message is None:
                message = self.get_message(starting=starting)
            if self.is_jupyter:
                # @NOTE: IPython isn't available in CI. This won't ever get invoked w/out IPython available though.
                from IPython.display import HTML, display # pyright: ignore[reportMissingImports]
                color_string = ""
                if color:
                    color_value = Color.parse(color)
                    rgb_tuple = color_value.get_truecolor()
                    rgb_hex = f"#{rgb_tuple[0]:02X}{rgb_tuple[1]:02X}{rgb_tuple[2]:02X}"
                    color_string = f"color: {rgb_hex};" if color is not None else ""
                content = HTML(f"<span style='font-family: monospace;{color_string}'>{message}</span>")
                if self.display is not None:
                    self.display.update(content)
                else:
                    self.display = display(content, display_id=True)
            else:
                if self.can_use_terminal_colors() and color is not None:
                    rich_message = f"[{color}]{message}"
                else:
                    rich_message = message
                rich_string = rich_str(rich_message)
                def width(word):
                    return sum(wcwidth(c) for c in word)
                diff = width(self.last_message) - width(rich_string)
                self.reset_cursor()
                # Use rich.print with lock protection
                output_file = file or sys.stdout
                rich.print(rich_message + (" " * diff), file=output_file, end="", flush=False)
                if output_file.isatty() or self.in_notebook:
                    output_file.flush()
                self.last_message = rich_string
    
    def can_use_terminal_colors(self):
        return not self.is_snowflake_notebook

    def update_messages(self, updater: dict[str, str]):
        if "message" in updater:
            self.message = updater["message"]
        if "finished_message" in updater:
            self.finished_message = updater["finished_message"]
        if "failed_message" in updater:
            self.failed_message = updater["failed_message"]
        self.update()

    def spinner_task(self):
        while self.busy and self.delay:
            self.update(color="magenta")
            time.sleep(self.delay) #type: ignore[union-attr] | we only call spinner_task if delay is not None anyway
            self.reset_cursor()

    def reset_cursor(self):
        if self.is_tty:
            # Clear the entire line and move cursor to beginning
            sys.stdout.write("\r\033[K")
        elif not self.is_jupyter:
            sys.stdout.write("\r")

    def __enter__(self):
        if self.leading_newline:
            rich.print()
        self.update(color="magenta", starting=True)
        # return control to the event loop briefly so stdout can be sure to flush:
        if self.delay:
            time.sleep(0.25)
        self.reset_cursor()
        if not self.delay:
            return self
        self.busy = True
        threading.Thread(target=self.spinner_task).start()
        return self

    def __exit__(self, exception, value, _):
        self.busy = False
        if exception is not None:
            if self.failed_message is not None:
                self.update(f"{self.failed_message} {value}", color="yellow", file=sys.stderr)
                # Use rich.print with explicit newline to ensure proper formatting
                rich.print(file=sys.stderr)
                return True
            return False
        if self.delay: # will be None for non-interactive environments
            time.sleep(self.delay)
        self.reset_cursor()
        if self.finished_message != "":
            final_message = f"▰▰▰▰ {self.finished_message}"
            self.update(final_message, color="green")
            # Use rich.print with explicit newline to ensure proper formatting
            rich.print()
        elif self.finished_message == "":
            self.update("")
            self.reset_cursor()
        if self.trailing_newline:
            rich.print()

class DebuggingSpan:
    span: debugging.Span
    def __init__(self, span_type: str):
        self.span_type = span_type
        self.span_attrs = {}

    def attrs(self, **kwargs):
        self.span_attrs = kwargs
        return self

    def __enter__(self):
        self.span = debugging.span_start(self.span_type, **self.span_attrs)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        debugging.span_end(self.span)


class SpanSpinner(Spinner):
    span: debugging.Span
    def __init__(self, span_type: str, *spinner_args, **spinner_kwargs):
        super().__init__(*spinner_args, **spinner_kwargs)
        self.span_type = span_type
        self.span_attrs = {}

    def attrs(self, **kwargs):
        self.span_attrs = kwargs
        return self

    def __enter__(self):
        self.span = debugging.span_start(self.span_type, **self.span_attrs)
        return super().__enter__()

    def __exit__(self, exc_type, exc_val, exc_tb):
        super().__exit__(exc_type, exc_val, exc_tb)
        debugging.span_end(self.span)



import logging
import uuid
import sys

from itertools import count
from relationalai.early_access.tests.logging import Capturer
from relationalai.early_access.metamodel import ir, types
from relationalai.early_access.builder import builder
from relationalai.clients.result_helpers import sort_data_frame_result
from relationalai.clients.util import IdentityParser
from relationalai.clients.snowflake import Provider as SFProvider
from relationalai import Provider
from typing import cast
from pathlib import Path

class SnapshotTestBase:
    """Base class for tests with snapshots."""

    provider:Provider = cast(SFProvider, Provider()) # type: ignore

    def run_snapshot_test(self, snapshot, script_path, db_schema=None, use_sql=False, use_lqp=False, match_internals=False):
        """Run a snapshot test.

        This method assumes content is already in the correct string format.
        Subclasses should handle any necessary compilation or conversion.
        """
        logger = logging.getLogger("pyrellogger")
        handler = Capturer()
        logger.addHandler(handler)

        # reset state across test runs
        self.reset_state()

        # Track which modules are already loaded
        before_modules = set(sys.modules.keys())
        unique_name = f"md{str(uuid.uuid4())[-12:]}"
        export_fqn = None
        try:
            overrides = {
                'dry_run': False,
                'model_suffix': f"_{unique_name}",
                'use_sql': use_sql,
                'reasoner.rule.use_lqp': use_lqp,
                'keep_model': False,
            }
            with builder.with_overrides(**overrides):
                with open(script_path, "r") as f:
                    # Compile with the correct filename
                    code = compile(f.read(), script_path, "exec")
                    # Execute the compiled code
                    if db_schema:
                        export_fqn = f"{db_schema}.{unique_name}"
                    exec_globals = {"EXPORT_TABLE": export_fqn}
                    exec(code, exec_globals)
        finally:
            # Remove any modules that were imported during exec()
            after_modules = set(sys.modules.keys())
            new_modules = after_modules - before_modules
            for mod in new_modules:
                sys.modules.pop(mod, None)
            # Remove handler
            logger.removeHandler(handler)
            # check if script created an export table
            export_exists = self.is_export_exists(export_fqn)
            # match snapshots
            self.assert_match_snapshots(snapshot, handler, export_exists, export_fqn, use_sql=use_sql, use_lqp=use_lqp, match_internals=match_internals)
            # cleanup resources created during a test run
            self.cleanup(export_exists, export_fqn)

    def assert_match_snapshots(self, snapshot, handler, export_exists=False, export_fqn=None, use_sql=False, use_lqp=False, match_internals=False):
        # compare model and queries only for internal e2e tests
        # todo: remove when e2e merged with unified tests
        if match_internals:
            if use_sql:
                extension = "sql"
            elif use_lqp:
                extension = "lqp"
            else:
                extension = "rel"

            for i, msg in enumerate(handler.model, start=1):
                snapshot.assert_match(str(msg["metamodel"]), f"model_{i}.metamodel")
                snapshot.assert_match(str(msg[extension]), f"model_{i}.{extension}")

            for i, msg in enumerate(handler.queries, start=1):
                snapshot.assert_match(str(msg["metamodel"]), f"query_{i}.metamodel")
                snapshot.assert_match(str(msg[extension]), f"query_{i}.{extension}")

            # The SQL compiler does not use ExtractNestedLogicals, so the type information is for a
            # different IR and we can't really check against what was generated by Rel and LQP
            if not use_sql:
                for i, msg in enumerate(handler.types, start=1):
                    snapshot.assert_match(str(msg["type_report"]), f"type_report_{i}.fish")

        for i, msg in enumerate(handler.results, start=1):
            if "results" in msg:
                snapshot.assert_match(msg["results"].to_csv(index=False).strip('"').strip('\n"'), f"result_{i}.csv")
        if export_fqn and export_exists:
            result_df = sort_data_frame_result(self.provider.sql(f"SELECT * FROM {export_fqn}", format="pandas"))
            snapshot.assert_match(result_df.to_csv(index=False).strip('"').strip('\n"'), "export_result.csv")

    def cleanup(self, export_exists=False, export_fqn=None):
        if export_fqn and export_exists :
            self.provider.sql(f"DROP TABLE IF EXISTS {export_fqn}")

    def is_export_exists(self, export_fqn=None):
        if export_fqn is None:
            return False
        database, schema, table, _ = IdentityParser(export_fqn, require_all_parts=True).to_list()
        return self.provider.sql(f"""
            SELECT EXISTS(
                SELECT 1
                FROM {database}.INFORMATION_SCHEMA.TABLES
                WHERE TABLE_SCHEMA = '{schema}'
                AND TABLE_NAME = '{table}'
            ) AS TABLE_EXISTS;
        """)[0]["TABLE_EXISTS"]

    @staticmethod
    def parent_tests(file: str):
        return SnapshotTestBase.discover_test_scripts(Path(file).parent)

    @staticmethod
    def discover_test_scripts(path: Path):
        scripts_dir = path / "tests"
        return sorted(scripts_dir.glob("**/*.py"))

    @staticmethod
    def get_id(p):
        p_name = p.parent.name
        return p.stem if p_name == "tests" else f"{p_name}_{p.stem}"

    def reset_state(self):
        """
        When we execute a pyrel program we accumulate some state, such as custom decimals, and
        we increase the counter for object ids. This function resets those counters and other
        state to a known baseline to ensure that test snapshots are consistent if we run the
        test alone vs in a suite of tests.
        """
        # reset the global id counters
        ir._global_id = count(10000)
        builder._global_id = count(10000)
        # Used to generate error source IDs
        builder.errors.ModelError.error_locations.clear()
        # caches of custom decimals
        for k in list(types._decimal_types.keys()):
            if types._decimal_types[k] != types.Decimal:
                del types._decimal_types[k]
        for k in list(builder.Concept.builtins):
            if k.startswith("Decimal(") and builder.Concept.builtins[k] != builder.Decimal:
                del builder.Concept.builtins[k]



"""
    Elementary IR relations.
"""
import sys

from . import ir, factory as f
from . import types

from typing import Optional

#
# Relations
#

# Comparators
def _comparator(name: str, input=True):
    overloads = [
        f.relation(name, [f.field("a", type, input), f.field("b", type, input)])
        for type in [types.Bool, types.Int64, types.Int128, types.Float, types.GenericDecimal, types.String, types.Date, types.DateTime, types.Hash, types.EntityTypeVar]
    ]
    return f.relation(name, [f.field("a", types.Any, input), f.field("b", types.Any, input)], overloads=overloads)

gt = _comparator(">")
gte = _comparator(">=")
lt = _comparator("<")
lte = _comparator("<=")
neq = _comparator("!=")
eq = _comparator("=", False)

# Arithmetic operators
def _binary_op(name: str, with_string=False, result_type: Optional[ir.Type]=None):
    overload_types = [types.Int64, types.Int128, types.Float, types.GenericDecimal]
    if with_string:
        overload_types.append(types.String)
    overloads = [
        f.relation(name, [
            f.input_field("a", type),
            f.input_field("b", type),
            f.field("c", result_type if result_type is not None else type)])
        for type in overload_types
    ]

    if with_string:
        return f.relation(name, [f.input_field("a", types.Any), f.input_field("b", types.Any), f.field("c", types.Any)], overloads=overloads)
    else:
        # If strings isn't added, then we're guaranteed to only have number overloads
        result_type = result_type if result_type is not None else types.Number
        return f.relation(name, [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", result_type)], overloads=overloads)

plus = _binary_op("+", with_string=True)
minus = _binary_op("-")
mul = _binary_op("*")
div = f.relation(
    "/",
    [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", types.Number)],
    overloads=[
        f.relation("/", [f.input_field("a", types.Int64), f.input_field("b", types.Int64), f.field("c", types.Float)]),
        f.relation("/", [f.input_field("a", types.Int128), f.input_field("b", types.Int128), f.field("c", types.Float)]),
        f.relation("/", [f.input_field("a", types.Float), f.input_field("b", types.Float), f.field("c", types.Float)]),
        f.relation("/", [f.input_field("a", types.GenericDecimal), f.input_field("b", types.GenericDecimal), f.field("c", types.GenericDecimal)]),
    ],
)
mod = _binary_op("%")
power = _binary_op("^")

trunc_div = f.relation(
    "//",
    [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", types.Number)],
    overloads=[
        f.relation("//", [f.input_field("a", types.Int64), f.input_field("b", types.Int64), f.field("c", types.Int64)]),
        f.relation("//", [f.input_field("a", types.Int128), f.input_field("b", types.Int128), f.field("c", types.Int128)]),
    ],
)

abs = f.relation(
    "abs",
    [f.input_field("a", types.Number), f.field("b", types.Number)],
    overloads=[
        f.relation("abs", [f.input_field("a", types.Int64), f.field("b", types.Int64)]),
        f.relation("abs", [f.input_field("a", types.Int128), f.field("b", types.Int128)]),
        f.relation("abs", [f.input_field("a", types.Float), f.field("b", types.Float)]),
        f.relation("abs", [f.input_field("a", types.GenericDecimal), f.field("b", types.GenericDecimal)]),
    ],
)

natural_log = f.relation(
    "natural_log",
    [f.input_field("a", types.Number), f.field("b", types.Number)],
    overloads=[
        f.relation("natural_log", [f.input_field("a", types.Int64), f.field("b", types.Float)]),
        f.relation("natural_log", [f.input_field("a", types.Int128), f.field("b", types.Float)]),
        f.relation("natural_log", [f.input_field("a", types.Float), f.field("b", types.Float)]),
        f.relation("natural_log", [f.input_field("a", types.GenericDecimal), f.field("b", types.GenericDecimal)]),

    ],
)

sqrt = f.relation(
    "sqrt",
    [f.input_field("a", types.Number), f.field("b", types.Number)],
    overloads=[
        f.relation("sqrt", [f.input_field("a", types.Int64), f.field("b", types.Float)]),
        f.relation("sqrt", [f.input_field("a", types.Int128), f.field("b", types.Float)]),
        f.relation("sqrt", [f.input_field("a", types.Float), f.field("b", types.Float)]),
        f.relation("sqrt", [f.input_field("a", types.GenericDecimal), f.field("b", types.GenericDecimal)]),
    ],
)

maximum = f.relation(
    "maximum",
    [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", types.Number)],
    overloads=[
        f.relation("maximum", [f.input_field("a", types.Int64), f.input_field("b", types.Int64), f.field("c", types.Int64)]),
        f.relation("maximum", [f.input_field("a", types.Int128), f.input_field("b", types.Int128), f.field("c", types.Int128)]),
        f.relation("maximum", [f.input_field("a", types.Float), f.input_field("b", types.Float), f.field("c", types.Float)]),
        f.relation("maximum", [f.input_field("a", types.GenericDecimal), f.input_field("b", types.GenericDecimal), f.field("c", types.GenericDecimal)]),
    ],
)

minimum = f.relation(
    "minimum",
    [f.input_field("a", types.Number), f.input_field("b", types.Number), f.field("c", types.Number)],
    overloads=[
        f.relation("minimum", [f.input_field("a", types.Int64), f.input_field("b", types.Int64), f.field("c", types.Int64)]),
        f.relation("minimum", [f.input_field("a", types.Int128), f.input_field("b", types.Int128), f.field("c", types.Int128)]),
        f.relation("minimum", [f.input_field("a", types.Float), f.input_field("b", types.Float), f.field("c", types.Float)]),
        f.relation("minimum", [f.input_field("a", types.GenericDecimal), f.input_field("b", types.GenericDecimal), f.field("c", types.GenericDecimal)]),
    ],
)

ceil = f.relation(
    "ceil",
    [f.input_field("a", types.Number), f.field("b", types.Number)],
    overloads=[
        f.relation("ceil", [f.input_field("a", types.Int64), f.field("b", types.Int64)]),
        f.relation("ceil", [f.input_field("a", types.Int128), f.field("b", types.Int128)]),
        f.relation("ceil", [f.input_field("a", types.Float), f.field("b", types.Float)]),
        f.relation("ceil", [f.input_field("a", types.GenericDecimal), f.field("b", types.GenericDecimal)]),
    ],
)

floor = f.relation(
    "floor",
    [f.input_field("a", types.Number), f.field("b", types.Number)],
    overloads=[
        f.relation("floor", [f.input_field("a", types.Int64), f.field("b", types.Int64)]),
        f.relation("floor", [f.input_field("a", types.Int128), f.field("b", types.Int128)]),
        f.relation("floor", [f.input_field("a", types.Float), f.field("b", types.Float)]),
        f.relation("floor", [f.input_field("a", types.GenericDecimal), f.field("b", types.GenericDecimal)]),
    ],
)

isinf = f.relation("isinf", [f.input_field("a", types.Float)])
isnan = f.relation("isnan", [f.input_field("a", types.Float)])

# Strings
concat = f.relation("concat", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.String)])
num_chars = f.relation("num_chars", [f.input_field("a", types.String), f.field("b", types.Int128)])
starts_with = f.relation("starts_with", [f.input_field("a", types.String), f.input_field("b", types.String)])
ends_with = f.relation("ends_with", [f.input_field("a", types.String), f.input_field("b", types.String)])
contains = f.relation("contains", [f.input_field("a", types.String), f.input_field("b", types.String)])
substring = f.relation("substring", [f.input_field("a", types.String), f.input_field("b", types.Int64), f.input_field("c", types.Int64), f.field("d", types.String)])
like_match = f.relation("like_match", [f.input_field("a", types.String), f.field("b", types.String)])
lower = f.relation("lower", [f.input_field("a", types.String), f.field("b", types.String)])
upper = f.relation("upper", [f.input_field("a", types.String), f.field("b", types.String)])
strip = f.relation("strip", [f.input_field("a", types.String), f.field("b", types.String)])

# Dates
date_format = f.relation("date_format", [f.input_field("a", types.Date), f.input_field("b", types.String), f.field("c", types.String)])
datetime_format = f.relation("datetime_format", [f.input_field("a", types.DateTime), f.input_field("b", types.String), f.input_field("c", types.String), f.field("d", types.String)])
date_year = f.relation("date_year", [f.input_field("a", types.Date), f.field("b", types.Int64)])
date_month = f.relation("date_month", [f.input_field("a", types.Date), f.field("b", types.Int64)])
date_day = f.relation("date_day", [f.input_field("a", types.Date), f.field("b", types.Int64)])
date_add = f.relation("date_add", [f.input_field("a", types.Date), f.input_field("b", types.Int64), f.field("c", types.Date)])
dates_period_days = f.relation("dates_period_days", [f.input_field("a", types.Date), f.input_field("b", types.Date), f.field("c", types.Int64)])
datetimes_period_milliseconds = f.relation("datetimes_period_milliseconds", [f.input_field("a", types.DateTime), f.input_field("b", types.DateTime), f.field("c", types.Int64)])
date_subtract = f.relation("date_subtract", [f.input_field("a", types.Date), f.input_field("b", types.Int64), f.field("c", types.Date)])
datetime_add = f.relation("datetime_add", [f.input_field("a", types.DateTime), f.input_field("b", types.Int64), f.field("c", types.DateTime)])
datetime_subtract = f.relation("datetime_subtract", [f.input_field("a", types.DateTime), f.input_field("b", types.Int64), f.field("c", types.DateTime)])

# Other
range = f.relation("range", [
    f.input_field("start", types.Int64),
    f.input_field("stop", types.Int64),
    f.input_field("step", types.Int64),
    f.field("result", types.Int64),
])

hash = f.relation("hash", [f.input_field("args", types.AnyList), f.field("hash", types.Hash)])

uuid_to_string = f.relation("uuid_to_string", [f.input_field("a", types.Hash), f.field("b", types.String)])

# Raw source code to be attached to the transaction, when the backend understands this language
raw_source = f.relation("raw_source", [f.input_field("lang", types.String), f.input_field("source", types.String)])

#
# Annotations
#

# indicates a relation is external to the system and, thus, backends should not rename or
# otherwise modify it
external = f.relation("external", [])
external_annotation = f.annotation(external, [])

# indicates an output is meant to be exported
export = f.relation("export", [])
export_annotation = f.annotation(export, [])

# indicates this relation is a concept population
concept_population = f.relation("concept_population", [])
concept_relation_annotation = f.annotation(concept_population, [])

# indicates this relation came in from CDC and will need to be shredded in Rel
from_cdc = f.relation("from_cdc", [])
from_cdc_annotation = f.annotation(from_cdc, [])

# indicates an = lookup that is from a cast operation for value types
from_cast = f.relation("from_cast", [])
from_cast_annotation = f.annotation(from_cast, [])

# indicates the original keys of an output (before they were replaced by a compound key)
output_keys = f.relation("output_keys", [])
output_keys_annotation = f.annotation(output_keys, [])

#
# Aggregations
#
def aggregation(name: str, params: list[ir.Field], overload_types: Optional[list[tuple[ir.Type, ...]]] = None):
    """Defines an aggregation, which is a Relation whose first 2 fields are a projection
    and a group, followed by the params."""
    fields = params
    overloads = []
    if overload_types:
        param_sets = []
        for ts in overload_types:
            param_sets.append([ir.Field(param.name, t, param.input) for param, t in zip(params, ts)])
        overloads = [
            aggregation(name, typed_params, overload_types=None)
            for typed_params in param_sets
        ]
    return f.relation(name, fields, overloads=overloads)

# concat = aggregation("concat", [
#     f.input_field("sep", types.String),
#     f.input_field("over", types.StringSet),
#     f.field("result", types.String)
# ])
# note that count does not need "over" because it counts the projection
count = aggregation("count", [
    f.field("result", types.Int128)
])
stats = aggregation("stats", [
    f.input_field("over", types.Number),
    f.field("std_dev", types.Number),
    f.field("mean", types.Number),
    f.field("median", types.Number),
])
sum = aggregation("sum", [
    f.input_field("over", types.Number),
    f.field("result", types.Number)
], overload_types=[
    (types.Int64, types.Int64),
    (types.Int128, types.Int128),
    (types.Float, types.Float),
    (types.GenericDecimal, types.GenericDecimal),
])
avg = aggregation("avg", [
    f.input_field("over", types.Number),
    f.field("result", types.Number)
], overload_types=[
    (types.Int64, types.Float), # nb. Float because Int / Int is Float
    (types.Int128, types.Float), # nb. Float because Int / Int is Float
    (types.Float, types.Float),
    (types.GenericDecimal, types.GenericDecimal),
])
max = aggregation("max", [
    f.input_field("over", types.Any),
    f.field("result", types.Any)
], overload_types=[
    (types.Int64, types.Int64),
    (types.Int128, types.Int128),
    (types.Float, types.Float),
    (types.GenericDecimal, types.GenericDecimal),
    (types.String, types.String),
    (types.Date, types.Date),
    (types.DateTime, types.DateTime),
    (types.EntityTypeVar, types.EntityTypeVar),
])
min = aggregation("min", [
    f.input_field("over", types.Any),
    f.field("result", types.Any)
], overload_types=[
    (types.Int64, types.Int64),
    (types.Int128, types.Int128),
    (types.Float, types.Float),
    (types.GenericDecimal, types.GenericDecimal),
    (types.String, types.String),
    (types.Date, types.Date),
    (types.DateTime, types.DateTime),
    (types.EntityTypeVar, types.EntityTypeVar),
])


#
# Pragmas
#
rule_reasoner_sem_vo = f.relation("rule_reasoner_sem_vo", [f.input_field("args", types.AnyList)])
rule_reasoner_phys_vo = f.relation("rule_reasoner_phys_vo", [f.input_field("args", types.AnyList)])


# TODO: these are Rel specific, should be moved from here
# Conversions
string = f.relation("string", [f.input_field("a", types.Any), f.field("b", types.String)])
parse_date = f.relation("parse_date", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.Date)])
parse_datetime = f.relation("parse_datetime", [f.input_field("a", types.String), f.input_field("b", types.String), f.field("c", types.DateTime)])
parse_decimal = f.relation("parse_decimal", [f.input_field("a", types.String), f.field("b", types.GenericDecimal)])
parse_int64 = f.relation("parse_int64", [f.input_field("a", types.String), f.field("b", types.Int64)])
parse_int128 = f.relation("parse_int128", [f.input_field("a", types.String), f.field("b", types.Int128)])

millisecond = f.relation("millisecond", [f.input_field("a", types.Int64), f.field("b", types.Int64)])
second = f.relation("second", [f.input_field("a", types.Int64), f.field("b", types.Int64)])
minute = f.relation("minute", [f.input_field("a", types.Int64), f.field("b", types.Int64)])
hour = f.relation("hour", [f.input_field("a", types.Int64), f.field("b", types.Int64)])
day = f.relation("day", [f.input_field("a", types.Int64), f.field("b", types.Int64)])
week = f.relation("week", [f.input_field("a", types.Int64), f.field("b", types.Int64)])
month = f.relation("month", [f.input_field("a", types.Int64), f.field("b", types.Int64)])
year = f.relation("year", [f.input_field("a", types.Int64), f.field("b", types.Int64)])

cast = f.relation(
    "cast",
    [
        f.input_field("to_type", types.Any),
        f.input_field("source",  types.Any),
        f.field("target",        types.Any)
    ],
    annos=[from_cast_annotation]
)

# Date construction with less overhead
construct_date = f.relation("construct_date", [f.input_field("year", types.Int64), f.input_field("month", types.Int64), f.input_field("day", types.Int64), f.field("date", types.Date)])
construct_date_from_datetime = f.relation("construct_date_from_datetime", [f.input_field("datetime", types.DateTime), f.input_field("timezone", types.String), f.field("date", types.Date)])
construct_datetime = f.relation("construct_datetime", [f.input_field("year", types.Int64), f.input_field("month", types.Int64), f.input_field("day", types.Int64), f.input_field("hour", types.Int64), f.input_field("minute", types.Int64), f.input_field("second", types.Int64), f.field("datetime", types.DateTime)])
construct_datetime_ms_tz = f.relation("construct_datetime_ms_tz", [f.input_field("year", types.Int64), f.input_field("month", types.Int64), f.input_field("day", types.Int64), f.input_field("hour", types.Int64), f.input_field("minute", types.Int64), f.input_field("second", types.Int64), f.input_field("milliseconds", types.Int64), f.input_field("timezone", types.String), f.field("datetime", types.DateTime)])

# Solver helpers
rel_primitive_solverlib_fo_appl = f.relation("rel_primitive_solverlib_fo_appl", [
    f.input_field("op", types.Int64),
    f.input_field("args", types.AnyList),
    f.field("result", types.String),
])
rel_primitive_solverlib_ho_appl = aggregation("rel_primitive_solverlib_ho_appl", [
    f.input_field("over", types.Any),
    f.field("op", types.Int64),
    f.field("result", types.String),
])
implies = f.relation("implies", [f.input_field("a", types.Bool), f.input_field("b", types.Bool)])
all_different = aggregation("all_different", [f.input_field("over", types.Any)])

#
# Public access to built-in relations
#

def is_builtin(r: ir.Relation):
    return r in builtin_relations or r in builtin_overloads

def is_annotation(r: ir.Relation):
    return r in builtin_annotations

def is_pragma(r: ir.Relation):
    return r in pragma_builtins

def _compute_builtin_relations() -> list[ir.Relation]:
    module = sys.modules[__name__]
    relations = []
    for name in dir(module):
        builtin = getattr(module, name)
        if isinstance(builtin, ir.Relation) and builtin not in builtin_annotations:
            relations.append(builtin)
    return relations

def _compute_builtin_overloads() -> list[ir.Relation]:
    module = sys.modules[__name__]
    overloads = []
    for name in dir(module):
        builtin = getattr(module, name)
        if isinstance(builtin, ir.Relation) and builtin not in builtin_annotations:
            if builtin.overloads:
                for overload in builtin.overloads:
                    if overload not in builtin_annotations:
                        overloads.append(overload)
    return overloads

# manually maintain the list of relations that are actually annotations
builtin_annotations = [external, export, concept_population, from_cdc, from_cast]
builtin_annotations_by_name = dict((r.name, r) for r in builtin_annotations)

builtin_relations = _compute_builtin_relations()
builtin_overloads = _compute_builtin_overloads()
builtin_relations_by_name = dict((r.name, r) for r in builtin_relations)

string_binary_builtins = [num_chars, starts_with, ends_with, contains, like_match, lower, upper, strip]

date_builtins = [date_year, date_month, date_day, date_add, date_subtract, datetime_add, datetime_subtract, dates_period_days, datetimes_period_milliseconds]

date_periods = [year, month, week, day, hour, minute, second, millisecond]

math_builtins = [abs, *abs.overloads, natural_log, *natural_log.overloads, sqrt, *sqrt.overloads,
                 maximum, *maximum.overloads, minimum, *minimum.overloads, mod, *mod.overloads,
                 power, *power.overloads, trunc_div, *trunc_div.overloads, ceil, *ceil.overloads,
                 floor, *floor.overloads]

pragma_builtins = [rule_reasoner_sem_vo, rule_reasoner_phys_vo]



from typing import cast, Optional

import relationalai as rai
from relationalai import Config
from relationalai.early_access.dsl.snow.common import TabularMetadata, ColumnMetadata, SchemaMetadata, \
    ForeignKey, ColumnRef


class Executor:
    def __init__(self, config: Optional[Config] = None):
        self._provider = cast(rai.clients.snowflake.Provider, rai.Provider(config=config))
        self._table_meta_cache = {}
        self._schema_fk_cache = {}

    def table_metadata(self, name: str):
        self._validate_table_name(name)
        if name in self._table_meta_cache:
            return self._table_meta_cache[name]
        # name comes in the form of `{database_name}.{schema_name}.{table_name}`
        database_nm, schema_nm, table_nm = name.upper().split(".")
        query = f"""
        SELECT TABLE_NAME, COLUMN_NAME, DATA_TYPE, IS_NULLABLE, NUMERIC_PRECISION, NUMERIC_PRECISION_RADIX, NUMERIC_SCALE
        FROM {database_nm}.INFORMATION_SCHEMA.COLUMNS
        WHERE TABLE_SCHEMA = '{schema_nm}' AND TABLE_NAME = '{table_nm}'
        ORDER BY TABLE_NAME, COLUMN_NAME, DATA_TYPE
        """
        # execute SQL
        rez = self._provider.sql(query)
        meta = self._parse_table_metadata(name, rez)
        # supplement with FKs
        schema_nm = f"{database_nm}.{schema_nm}".lower()
        self._supplement_with_foreign_keys(meta, schema_nm)
        # cache before returning
        self._table_meta_cache[name] = meta
        return meta

    def _supplement_with_foreign_keys(self, tb_meta, schema_name: str):
        schema_fk_meta = self.schema_foreign_keys(schema_name)
        # suboptimal but works for now
        for fk in schema_fk_meta.foreign_keys:
            if tb_meta.name == fk.source_columns[0].table:
                tb_meta.foreign_keys.add(fk)

    @staticmethod
    def _parse_table_metadata(name, result):
        meta = TabularMetadata(name=name)
        for row in result:
            attr = ColumnMetadata(
                name=row['COLUMN_NAME'],
                datatype=row['DATA_TYPE'],
                is_nullable=row['IS_NULLABLE'] == 'YES',
                numeric_precision=row['NUMERIC_PRECISION'],
                numeric_precision_radix=row['NUMERIC_PRECISION_RADIX'],
                numeric_scale=row['NUMERIC_SCALE']
            )
            meta.columns.append(attr)
        return meta

    @staticmethod
    def _validate_table_name(name):
        # check if the table name is in the form of `{database_name}.{schema_name}.{table_name}`
        if len(name.split('.')) != 3:
            raise ValueError(f'Table name {name} is not in the correct format')

    def schema_foreign_keys(self, name):
        self._validate_schema_name(name)
        if name in self._schema_fk_cache:
            return self._schema_fk_cache[name]

        # name comes in the form of `{database_name}.{schema_name}`
        query = f"""
        SHOW IMPORTED KEYS IN SCHEMA {name};
        """
        rez = self._provider.sql(query)
        meta = self._parse_schema_foreign_keys(name, rez)
        self._schema_fk_cache[name] = meta
        return meta

    @staticmethod
    def _parse_schema_foreign_keys(name, results):
        metadata = SchemaMetadata(name=name)
        for row in results:
            fk_table_nm = f'{row["fk_database_name"]}.{row["fk_schema_name"]}.{row["fk_table_name"]}'.lower()
            pk_table_nm = f'{row["pk_database_name"]}.{row["pk_schema_name"]}.{row["pk_table_name"]}'.lower()
            # get columns refs, since FK may be composite, we need to handle multiple columns
            # TODO: implement multi-col FK support
            source_cols = [ColumnRef(fk_table_nm, row['fk_column_name'])]
            target_cols = [ColumnRef(pk_table_nm, row['pk_column_name'])]
            fk = ForeignKey(
                name=row['fk_name'],
                source_columns=source_cols,
                target_columns=target_cols
            )
            metadata.foreign_keys.append(fk)
        return metadata

    @staticmethod
    def _validate_schema_name(name):
        # check if the schema name is in the form of `{database_name}.{schema_name}`
        if len(name.split('.')) != 2:
            raise ValueError(f'Schema name {name} is not in the correct format')

    def provider(self):
        return self._provider
